from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.linux.proc import ProcPlugin
from dissect.target.plugins.os.unix.linux.sockets import NetSocketPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_sockets_plugin(target_linux_users: Target, fs_linux_proc_sockets: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)
    target_linux_users.add_plugin(NetSocketPlugin)
    results = list(target_linux_users.sockets())
    assert len(results) == 24


def test_tcp(target_linux_users: Target, fs_linux_proc_sockets: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)
    target_linux_users.add_plugin(NetSocketPlugin)

    results = list(target_linux_users.sockets.tcp())

    assert len(results) == 7
    assert results[0].local_ip == "0.0.0.0"
    assert results[0].local_port == 22
    assert results[0].remote_ip == "0.0.0.0"
    assert results[0].remote_port == 0

    assert results[2].local_ip == "127.0.0.1"
    assert results[2].local_port == 631
    assert results[2].remote_ip == "0.0.0.0"
    assert results[2].remote_port == 0

    assert results[3].local_ip == "172.16.64.136"
    assert results[3].local_port == 22
    assert results[3].remote_ip == "172.16.64.1"
    assert results[3].remote_port == 49442

    assert results[4].local_ip == "::"
    assert results[4].local_port == 22
    assert results[4].remote_ip == "::"
    assert results[4].remote_port == 0

    assert results[6].local_ip == "::1"
    assert results[6].local_port == 631
    assert results[6].remote_ip == "::"
    assert results[6].remote_port == 0

    for result in results:
        assert result.protocol in ("tcp", "tcp6")
        assert result.owner == "root"
        assert result.pid in (0, 1, 2, 1337)
        assert result.inode in (1337, 1338, 0)
        assert result.cmdline in ("", None, "acquire -p full --proc", "test cmdline")


def test_udp(target_linux_users: Target, fs_linux_proc_sockets: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)
    target_linux_users.add_plugin(NetSocketPlugin)
    results = list(target_linux_users.sockets.udp())

    assert len(results) == 7
    assert results[0].local_ip == "172.16.64.136"
    assert results[0].local_port == 68
    assert results[0].remote_ip == "172.16.64.254"
    assert results[0].remote_port == 67

    assert results[2].local_ip == "0.0.0.0"
    assert results[2].local_port == 58569
    assert results[2].remote_ip == "0.0.0.0"
    assert results[2].remote_port == 0

    assert results[3].local_ip == "0.0.0.0"
    assert results[3].local_port == 5353
    assert results[3].remote_ip == "0.0.0.0"
    assert results[3].remote_port == 0

    assert results[4].local_ip == "::"
    assert results[4].local_port == 59613
    assert results[4].remote_ip == "::"
    assert results[4].remote_port == 0

    for result in results:
        assert result.protocol in ("udp", "udp6")
        assert result.owner in ("root", "110")
        assert result.pid in (0, 1, 2, 3, 1337)
        assert result.inode in (1337, 1338, 1339, 0)
        assert result.cmdline in ("", None, "acquire -p full --proc", "sshd", "test cmdline")


def test_raw(target_linux_users: Target, fs_linux_proc_sockets: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)
    target_linux_users.add_plugin(NetSocketPlugin)
    results = list(target_linux_users.sockets.raw())

    assert len(results) == 4
    assert results[0].local_ip == "0.0.0.0"
    assert results[0].local_port == 253
    assert results[0].remote_ip == "0.0.0.0"
    assert results[0].remote_port == 0

    assert results[1].local_ip == "0.0.0.0"
    assert results[1].local_port == 253
    assert results[1].remote_ip == "0.0.0.0"
    assert results[1].remote_port == 0

    for result in results:
        assert result.protocol in ("raw", "raw6")
        assert result.owner == "root"
        assert result.pid in (1, 1337)
        assert result.inode == (1337)
        assert result.cmdline in ("acquire -p full --proc", "test cmdline")


def test_packet(target_linux_users: Target, fs_linux_proc_sockets: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)
    target_linux_users.add_plugin(NetSocketPlugin)
    results = list(target_linux_users.sockets.packet())

    assert len(results) == 2

    for result in results:
        assert result.ref == 3
        assert result.protocol == "packet"
        assert result.protocol_type == "ETH_P_ALL"
        assert result.socket_type == 3
        assert result.sk == 2174716005
        assert result.cmdline in ("acquire -p full --proc", "test cmdline")
        assert result.pid in (1, 1337)
        assert result.uid == 0
        assert result.owner == "root"


def test_unix(target_linux_users: Target, fs_linux_proc_sockets: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)
    target_linux_users.add_plugin(NetSocketPlugin)
    results = list(target_linux_users.sockets.unix())

    assert len(results) == 4
    for result in results:
        assert result.ref in (2, 3)
        assert result.protocol == "unix"
        assert result.type == "STREAM"
        assert result.state in ("LISTENING", "CONNECTED")
        assert result.socket_flags in ("00010000", "00000000")

        assert result.path in (
            "/run/systemd/private",
            None,
            "/run/systemd/io.system.ManagedOOM",
            "@/tmp/dbus-YLq1FHVh",
        )



from __future__ import annotations

import pytest

from dissect.target.helpers.localeutil import normalize_language, normalize_timezone


def test_helpers_localeutil_normalize_timezone() -> None:
    assert normalize_timezone("W. Europe Standard Time") == "Europe/Berlin"
    assert normalize_timezone("GMT Standard Time") == "Europe/London"
    assert normalize_timezone("UTC") == "UTC"


@pytest.mark.parametrize(
    ("language_input", "expected_output"),
    [
        ("en_US.UTF-8", "en_US"),
        ("nl", "nl_NL"),
        ("en", "en_US"),
        ("en-US", "en_US"),
    ],
)
def test_helpers_localeutil_normalize_language(language_input: str, expected_output: str) -> None:
    """Test if we normalize languages to ISO-3166 correctly."""

    assert normalize_language(language_input) == expected_output



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator


class PodmanChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from Podman overlayfs containers."""

    __type__ = "podman"

    def check_compatible(self) -> None:
        if not self.target.has_function("podman"):
            raise UnsupportedPluginError("No Podman install(s) found on target")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for container in self.target.podman.containers():
            if container.mount_path:
                yield ChildTargetRecord(
                    type=self.__type__,
                    path=container.mount_path,
                    _target=self.target,
                )



from __future__ import annotations

import pytest

from dissect.target.loaders.itunes import translate_file_path


@pytest.mark.parametrize(
    ("domain", "relative_path", "result"),
    [
        pytest.param(
            "HomeDomain",
            "Library/Safari/SafariTabs.db",
            "/private/var/mobile/Library/Safari/SafariTabs.db",
            id="HomeDomain",
        ),
        pytest.param("ProtectedDomain", "", "ProtectedDomain", id="ProtectedDomain"),
        pytest.param(
            "AppDomain-com.apple.freeform",
            "",
            "/private/var/mobile/Containers/Data/Application/com.apple.freeform",
            id="AppDomain",
        ),
    ],
)
def test_translate_file_path(domain: str, relative_path: str, result: str) -> None:
    assert translate_file_path(domain, relative_path) == result



from __future__ import annotations

import re
import struct
from typing import TYPE_CHECKING

from dissect.util.ts import wintimestamp

from dissect.target.exceptions import (
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
    UnsupportedPluginError,
)
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import VirtualKey

UsbRegistryRecord = TargetRecordDescriptor(
    "windows/registry/usb",
    [
        ("string", "type"),
        ("string", "serial"),
        ("string", "container_id"),
        ("string", "vendor"),
        ("string", "product"),
        ("string", "revision"),
        ("string", "friendly_name"),
        ("datetime", "first_insert"),
        ("datetime", "first_install"),
        ("datetime", "last_insert"),
        ("datetime", "last_removal"),
        ("string[]", "volumes"),
        ("string[]", "mounts"),
        ("string[]", "users"),
        ("path", "source"),
    ],
)

USB_DEVICE_PROPERTY_KEYS = {
    "first_install": ("0064", "00000064"),  # Windows 7 and higher. USB device first install date
    "first_insert": ("0065", "00000065"),  # Windows 7 and higher. USB device first insert date.
    "last_insert": ("0066", "00000066"),  # Windows 8 and higher. USB device last insert date.
    "last_removal": ("0067", "00000067"),  # Windows 8 and higer. USB device last removal date.
}

RE_DEVICE_NAME = re.compile(r"^(?P<type>.+?)&Ven_(?P<vendor>.+?)&Prod_(?P<product>.+?)(&Rev_(?P<revision>.+?))?$")


class UsbPlugin(Plugin):
    """Windows USB history plugin.

    Parses Windows registry data about attached USB devices. Does not parse EVTX EventIDs
    or ``C:\\Windows\\inf\\setupapi(.dev).log``.

    To get a full picture of the USB history on a Windows machine, you should parse the
    relevant EventIDs using the evtx plugin. For more research on event log USB forensics, see:

        - https://www.researchgate.net/publication/318514858_USB_Storage_Device_Forensics_for_Windows_10
        - https://dfir.pubpub.org/pub/h78di10n/release/2
        - https://www.senturean.com/posts/19_08_03_usb_storage_forensics_1/#1-system-events

    References:
        - https://hatsoffsecurity.com/2014/06/05/usb-forensics-pt-1-serial-number/
        - http://www.swiftforensics.com/2013/11/windows-8-new-registry-artifacts-part-1.html
        - https://www.sans.org/blog/the-truth-about-usb-device-serial-numbers/
    """

    # Stores history of mounted USB devices
    USB_STOR = "HKLM\\SYSTEM\\CurrentControlSet\\Enum\\USBSTOR"

    # Stores the relation between a USB container_id and the FriendlyName of mounted volume(s) (Windows 7 and up)
    PORTABLE_DEVICES = "HKLM\\SOFTWARE\\Microsoft\\Windows Portable Devices\\Devices"

    # Stores the most recent mapping of a mount letter and a container_id
    MOUNT_LETTER_MAP = "HKLM\\SYSTEM\\MountedDevices"

    # User history of mount points accesses in explorer.exe
    USER_MOUNTS = "HKCU\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Mountpoints2"

    # Other artifacts we currently do not parse:
    # - "sysvol\Windows\inf\setupapi(.dev).log"
    # - "HKLM\\SYSTEM\\CurrentControlSet\\Enum\\USB"
    # - "HKLM\\SYSTEM\\CurrentControlSet\\Enum\\HID"
    # - "HKLM\\SYSTEM\\CurrentControlSet\\Enum\\SCSI"
    # - "HKLM\\SYSTEM\\CurrentControlSet\\Control\\DeviceContainers"
    # - "SOFTWARE\Microsoft\Windows NT\CurrentVersion\EMDMgmt"

    def check_compatible(self) -> None:
        if not list(self.target.registry.keys(self.USB_STOR)):
            raise UnsupportedPluginError(f"Registry key not found: {self.USB_STOR}")

    @export(record=UsbRegistryRecord)
    def usb(self) -> Iterator[UsbRegistryRecord]:
        """Yields information about (historically) attached USB storage devices on Windows.

        Uses the registry to find information about USB storage devices that have been attached to the system.
        Also tries to find the past volume name and mount letters of the USB device and what user(s) interacted
        with them using ``explorer.exe``.
        """

        for key in self.target.registry.keys(self.USB_STOR):
            for usb_type in key.subkeys():
                try:
                    device_info = parse_device_name(usb_type.name)
                except ValueError:
                    self.target.log.warning("Unable to parse USB device name: %s", usb_type.name)
                    device_info = {"type": None, "vendor": None, "product": None, "revision": None}

                for usb_device in usb_type.subkeys():
                    serial = usb_device.name
                    friendly_name = None
                    container_id = None
                    timestamps = {
                        "first_install": None,
                        "first_insert": None,
                        "last_insert": None,
                        "last_removal": None,
                    }

                    try:
                        friendly_name = usb_device.value("FriendlyName").value
                    except RegistryValueNotFoundError:
                        self.target.log.warning("No FriendlyName for USB with serial: %s", serial)

                    try:
                        container_id = usb_device.value("ContainerID").value
                    except RegistryValueNotFoundError:
                        self.target.log.warning("No ContainerID for USB with serial: %s", serial)

                    try:
                        timestamps = unpack_timestamps(usb_device.subkey("Properties"))
                    except RegistryValueNotFoundError as e:
                        self.target.log.warning("Unable to parse USBSTOR registry properties for serial: %s", serial)
                        self.target.log.debug("", exc_info=e)

                    # We can check if any HKCU hive(s) are populated with the Volume GUID of the USB storage device.
                    # If a user has interacted with the mounted volume using explorer.exe we will get a match.
                    volumes = list(self.find_volumes(serial))
                    mounts = list(self.find_mounts(serial))
                    users = [
                        u.user.name for u in self.find_users([m[10:] for m in mounts if m.startswith("\\??\\Volume{")])
                    ]

                    yield UsbRegistryRecord(
                        friendly_name=friendly_name,
                        serial=serial,
                        container_id=container_id,
                        **device_info,
                        **timestamps,
                        volumes=volumes,
                        mounts=mounts,
                        users=users,
                        source=self.USB_STOR,
                        _target=self.target,
                    )

    def find_volumes(self, serial: str) -> Iterator[str]:
        """Attempts to find mounted volume names for the given serial."""
        serial = serial.lower()
        try:
            for device in self.target.registry.key(self.PORTABLE_DEVICES).subkeys():
                if serial in device.name.lower():
                    yield device.value("FriendlyName").value
        except RegistryKeyNotFoundError:
            pass

    def find_mounts(self, serial: str) -> Iterator[str]:
        """Attempts to find drive letters the given serial has been mounted on."""
        serial = serial.lower()
        try:
            for mount in self.target.registry.key(self.MOUNT_LETTER_MAP).values():
                try:
                    if serial in mount.value.decode("utf-16-le").lower():
                        yield mount.name.replace("\\DosDevices\\", "")
                except UnicodeDecodeError:  # noqa: PERF203
                    pass
        except RegistryKeyNotFoundError:
            pass

    def find_users(self, volume_guids: list[str]) -> Iterator[str]:
        """Attempt to find Windows users that have interacted with the given volume GUIDs."""

        for volume_guid in volume_guids:
            try:
                for key in self.target.registry.key(self.USER_MOUNTS + "\\" + volume_guid):
                    yield self.target.registry.get_user_details(key)
            except RegistryKeyNotFoundError:  # noqa: PERF203
                pass


def unpack_timestamps(usb_reg_properties: VirtualKey) -> dict[str, int]:
    """Unpack relevant Windows timestamps from the provided USB registry properties subkey.

    Args:
        usb_reg_properties: A registry object with USB properties.

    Returns:
        A dict containing parsed timestamps within passed registry object.
    """

    usb_reg_properties = usb_reg_properties.subkey("{83da6326-97a6-4088-9453-a1923f573b29}")
    timestamps = {}

    for device_property, usbstor_values in USB_DEVICE_PROPERTY_KEYS.items():
        for usb_val in usbstor_values:
            if usb_val in [x.name for x in usb_reg_properties.subkeys()]:
                version_key = usb_reg_properties.subkey(usb_val)
                if "00000000" in version_key.subkeys():
                    data_value = version_key.subkey("00000000").value("Data").value
                else:
                    data_value = version_key.value("(Default)").value
                timestamps[device_property] = wintimestamp(struct.unpack("<Q", data_value)[0])
                break
            timestamps[device_property] = None
    return timestamps


def parse_device_name(device_name: str) -> dict[str, str]:
    """Parse a registry device name into components."""

    match = RE_DEVICE_NAME.match(device_name)
    if not match:
        raise ValueError(f"Unable to parse USB device name: {device_name}")

    return match.groupdict()



from __future__ import annotations

from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, NamedTuple, TypeVar

from dissect.target.exceptions import Error
from dissect.target.helpers.nfs.nfs3 import (
    CookieVerf,
    DirOpArgs,
    EntryPlus,
    FileAttributes,
    FileHandle,
    GetAttrProc,
    LookupProc,
    LookupResult,
    NfsStat,
    ReadDirPlusParams,
    ReadDirPlusProc,
    ReadFileProc,
    ReadLinkProc,
    ReadParams,
)
from dissect.target.helpers.nfs.serializer import (
    DirOpArgs3Serializer,
    FileAttributesSerializer,
    Lookup3ResultDeserializer,
    Read3ArgsSerializer,
    Read3ResultDeserializer,
    ReadDirPlusParamsSerializer,
    ReadDirPlusResultDeserializer,
    ReadLink3ResultDeserializer,
)
from dissect.target.helpers.nfs.serializer import (
    ResultDeserializer as NfsResultDeserializer,
)
from dissect.target.helpers.sunrpc.client import AbstractClient as SunRpcAbstractClient
from dissect.target.helpers.sunrpc.client import AuthScheme, LocalPortPolicy
from dissect.target.helpers.sunrpc.client import Client as SunRpcClient
from dissect.target.helpers.sunrpc.serializer import OpaqueVarLengthSerializer

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import TracebackType

    from typing_extensions import Self

Credentials = TypeVar("Credentials")
Verifier = TypeVar("Verifier")


class NfsError(Error):
    def __init__(self, message: str, nfsstat: NfsStat):
        super().__init__(message)
        self.message = message
        self.nfsstat = nfsstat

    def __str__(self) -> str:
        return f"{self.__class__.__name__} (NfsStat: {self.nfsstat}): {self.message}"


class ReadDirResult(NamedTuple):
    dir_attributes: FileAttributes | None
    entries: list[EntryPlus]


class Client(AbstractContextManager):
    DIR_COUNT = 4096  # See https://datatracker.ietf.org/doc/html/rfc1813#section-3.3.17
    MAX_COUNT = 32768
    READ_CHUNK_SIZE = 1024 * 1024

    def __init__(self, rpc_client: SunRpcAbstractClient):
        self._rpc_client = rpc_client

    def __enter__(self) -> Self:
        """Return ``self`` upon entering the runtime context."""
        return self  # type: Necessary for type checker

    def __exit__(self, _: type[BaseException] | None, __: BaseException | None, ___: TracebackType | None) -> bool:
        self.close()
        return False  # Reraise exceptions

    @classmethod
    def connect(
        cls,
        hostname: str,
        port: int,
        auth: AuthScheme[Credentials, Verifier],
        local_port: int | LocalPortPolicy = 0,
        timeout_in_seconds: float | None = 5.0,
    ) -> Client:
        """Connect to a NFS server.

        Args:
            hostname: The remote hostname.
            port: The remote port.
            auth: The authentication scheme.
            local_port: The local port to bind to.
                If equal to ``LocalPortPolicy.PRIVILEGED`` or -1, bind to the first free privileged port.
                If equal to ``LocalPortPolicy.ANY`` or 0, bind to any free port.
                Otherwise, bind to the specified port.
            timeout_in_seconds: The timeout for making the connection.
        """

        rpc_client = SunRpcClient.connect(hostname, port, auth, local_port, timeout_in_seconds)
        return Client(rpc_client)

    def rebind_auth(self, auth: AuthScheme[Credentials, Verifier]) -> None:
        """Change the authentication scheme of the underlying sunrpc client."""

        self._rpc_client = self._rpc_client.rebind_auth(auth)

    def readdir(self, dir: FileHandle) -> ReadDirResult | NfsStat:
        """Read the contents of a directory, including file attributes."""

        entries = list[EntryPlus]()
        cookie = 0
        cookieverf = CookieVerf(b"\x00")
        read_deserializer = NfsResultDeserializer(ReadDirPlusResultDeserializer())

        # Multiple calls might be needed to read the entire directory
        while True:
            args = ReadDirPlusParams(dir, cookie, cookieverf, dir_count=self.DIR_COUNT, max_count=self.MAX_COUNT)

            result = self._rpc_client.call(ReadDirPlusProc, args, ReadDirPlusParamsSerializer(), read_deserializer)
            if isinstance(result, NfsStat):
                raise NfsError("Failed to read dir", result)

            entries += result.entries
            if result.eof or len(result.entries) == 0:
                return ReadDirResult(result.dir_attributes, entries)

            cookie = result.entries[-1].cookie
            cookieverf = result.cookieverf

    def readfile(self, handle: FileHandle, offset: int = 0, size: int = -1) -> Iterator[bytes]:
        """Read a file by its file handle."""
        bytes_left = size
        read_deserializer = NfsResultDeserializer(Read3ResultDeserializer())
        while size == -1 or bytes_left > 0:
            count = self.READ_CHUNK_SIZE if size == -1 else min(self.READ_CHUNK_SIZE, size)
            params = ReadParams(handle, offset, count)
            result = self._rpc_client.call(ReadFileProc, params, Read3ArgsSerializer(), read_deserializer)
            if isinstance(result, NfsStat):
                raise NfsError("Failed to read file", result)
            yield result.data
            if result.eof:
                return
            offset += result.count
            bytes_left -= result.count

    def lookup(self, name: str, parent: FileHandle) -> LookupResult:
        """Lookup a file by name in a directory."""

        args = DirOpArgs(parent, name)
        lookup_deserializer = NfsResultDeserializer(Lookup3ResultDeserializer())
        result = self._rpc_client.call(LookupProc, args, DirOpArgs3Serializer(), lookup_deserializer)
        if isinstance(result, NfsStat):
            raise NfsError("Failed to lookup file", result)

        return result

    def getattr(self, handle: FileHandle) -> FileAttributes:
        """Get the attributes of a file by its file handle."""

        attr_deserializer = NfsResultDeserializer(FileAttributesSerializer())
        result = self._rpc_client.call(GetAttrProc, handle.opaque, OpaqueVarLengthSerializer(), attr_deserializer)
        if isinstance(result, NfsStat):
            raise NfsError("Failed to get attributes", result)

        return result

    def readlink(self, handle: FileHandle) -> str:
        """Read the target of a symlink by its file handle."""
        link_deserializer = NfsResultDeserializer(ReadLink3ResultDeserializer())
        result = self._rpc_client.call(ReadLinkProc, handle.opaque, OpaqueVarLengthSerializer(), link_deserializer)
        if isinstance(result, NfsStat):
            raise NfsError("Failed to read link", result)

        return result.target

    def close(self) -> None:
        self._rpc_client.close()

    def __del__(self):
        self.close()



from __future__ import annotations

import json
import re
from datetime import datetime
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.plugin import export
from dissect.target.plugins.apps.webserver.webserver import (
    WebserverAccessLogRecord,
    WebserverErrorLogRecord,
    WebserverHostRecord,
    WebserverPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

# Reference: https://nginx/org/en/docs/http/ngx_http_log_module.html#log_format
RE_ACCESS_LOG = re.compile(
    r"""
        (?P<remote_ip>.*?)\s-\s(?P<remote_user>.*?)
        \s
        \[(?P<datetime>\d{2}\/[A-Za-z]{3}\/\d{4}:\d{2}:\d{2}:\d{2}\s(\+|\-)\d{4})\]
        \s
        \"(?P<method>.*?)\s(?P<uri>.*?)\s?(?P<protocol>HTTP\/.*?)?\"
        \s
        (?P<status_code>\d{3})
        \s
        (?P<bytes_sent>-|\d+)
        \s
        (["](?P<referer>(\-)|(.+))["])
        \s
        \"(?P<useragent>.*?)\"
    """,
    re.VERBOSE,
)

# Reference: https://github.com/nginx/nginx/blob/master/src/core/ngx_log.c
RE_ERROR_LOG = re.compile(
    r"""
        (?P<ts>\d{4}\/\d{2}\/\d{2}\s\d{2}:\d{2}:\d{2})      # YYYY/MM/DD HH:MM:SS
        \s
        \[(?P<level>\S+)\]
        \s
        (?P<pid>\d+)\#(?P<tid>\d+)\:                        # 12345#12345:
        \s
        (?P<message>.+)
    """,
    re.VERBOSE,
)

# Reference: https://nginx.org/en/docs/http/ngx_http_log_module.html#access_log
RE_ACCESS_LOG_DIRECTIVE = re.compile(
    r"""
        (?:\#\s+?)?                     # optionally include disabled directives
        access_log\s+
        (?P<path>[^\s\;]+)
        (?:\s(?P<format>[^\s\;]+))?     # capture format and ignore other arguments
    """,
    re.VERBOSE,
)

# Reference: https://nginx.org/en/docs/ngx_core_module.html#error_log
RE_ERROR_LOG_DIRECTIVE = re.compile(
    r"""
        (?:\#\s+?)?                     # optionally include disabled directives
        error_log\s+
        (?P<path>[^\s\;]+)
        (?:\s(?P<level>[^\s\;]+))?      # capture optional level
    """,
    re.VERBOSE,
)

# Reference: https://nginx.org/en/docs/ngx_core_module.html#include
RE_INCLUDE_DIRECTIVE = re.compile(r"[\s#]*include\s+(?P<path>[^\s\;]+)")


class NginxPlugin(WebserverPlugin):
    """NGINX webserver plugin."""

    __namespace__ = "nginx"

    DEFAULT_LOG_DIRS = (
        "/var/log/nginx",
        "/var/log",
    )

    ACCESS_LOG_NAMES = ("access.log",)
    ERROR_LOG_NAMES = ("error.log",)

    DEFAULT_CONFIG_PATHS = (
        "/etc/nginx/nginx.conf",
        "/etc/nginx/sites-available/*.conf",
        "/etc/nginx/sites-enabled/*.conf",
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self.access_paths = set()
        self.error_paths = set()
        self.host_paths = set()

        self.find_logs()

    def check_compatible(self) -> None:
        if not self.access_paths and not self.error_paths and not self.host_paths:
            raise UnsupportedPluginError("No NGINX log or config files found on target")

    def find_logs(self) -> None:
        # Add any well known default NGINX log locations
        for log_dir in self.DEFAULT_LOG_DIRS:
            log_dir = self.target.fs.path(log_dir)
            for log_name in self.ACCESS_LOG_NAMES:
                self.access_paths.update(log_dir.glob(f"{log_name}*"))
            for log_name in self.ERROR_LOG_NAMES:
                self.error_paths.update(log_dir.glob(f"{log_name}*"))

        # Check for custom paths in NGINX install config
        for config_file in self.DEFAULT_CONFIG_PATHS:
            if "*" in config_file:
                base, _, glob = config_file.partition("*")
                for f in self.target.fs.path(base).rglob(f"*{glob}"):
                    self.parse_config(f)

            elif (config_file := self.target.fs.path(config_file)).exists():
                self.parse_config(config_file)

    def parse_config(self, path: Path, seen: set[Path] | None = None) -> None:
        """Parse the given NGINX ``.conf`` file for ``access_log``, ``error_log`` and ``include`` directives."""

        seen = set() if seen is None else seen

        if path in seen:
            self.target.log.warning("Detected recursion in NGINX configuration, file already parsed: %s", path)
            return

        seen.add(path)

        if not path.is_file():
            self.target.log.warning("File %s does not exist on target", path)
            return

        for line in path.open("rt"):
            if not (line := line.strip()):
                continue

            if "access_log " in line:
                if access_log := RE_ACCESS_LOG_DIRECTIVE.search(line):
                    access_log = self.target.fs.path(access_log["path"])
                    self.access_paths.update(access_log.parent.glob(f"{access_log.name}*"))
                else:
                    self.target.log.warning("Unable to parse nginx access_log line %r in %s", line, path)

            elif "error_log " in line:
                if error_log := RE_ERROR_LOG_DIRECTIVE.search(line):
                    error_log = self.target.fs.path(error_log["path"])
                    self.error_paths.update(error_log.parent.glob(f"{error_log.name}*"))
                else:
                    self.target.log.warning("Unable to parse NGINX error_log line %r in %s", line, path)

            elif "server {" in line:
                self.host_paths.add(path)

            elif "include " in line:
                if match := RE_INCLUDE_DIRECTIVE.search(line):
                    path_str: str = match.groupdict().get("path")

                    if "*" in path_str:
                        base, _, glob = path_str.partition("*")
                        include_paths = self.target.fs.path(base).rglob(f"*{glob}")
                    else:
                        include_paths = [self.target.fs.path(path_str)]

                    for include_path in include_paths:
                        if include_path.is_absolute():
                            self.parse_config(include_path)
                        else:
                            include_path = self.target.fs.path(path.parent).joinpath(include_path)
                            self.parse_config(include_path)
                else:
                    self.target.log.warning("Unable to parse NGINX include line %r in %s", line, path)

    @export(record=WebserverAccessLogRecord)
    def access(self) -> Iterator[WebserverAccessLogRecord]:
        """Return contents of NGINX access log files in unified ``WebserverAccessLogRecord`` format.

        References:
            - https://docs.nginx.com/nginx/admin-guide/monitoring/logging/#access_log
            - http://nginx.org/en/docs/http/ngx_http_log_module.html#log_format
        """
        for path in self.access_paths:
            path = path.resolve(strict=True)
            if not path.is_file():
                self.target.log.warning("NGINX log file configured but could not be found (dead symlink?): %s", path)
                continue

            for line in open_decompress(path, "rt"):
                if not (line := line.strip()):
                    continue

                log: dict[str, str] = {}

                if line[0:2] == '{"':
                    try:
                        log = parse_json_line(line)
                    except ValueError as e:
                        self.target.log.warning("Could not parse NGINX JSON log line %r in %s", line, path)
                        self.target.log.debug("", exc_info=e)
                        continue

                elif match := RE_ACCESS_LOG.search(line):
                    log = match.groupdict()

                else:
                    self.target.log.warning("Could not match NGINX format for log line %r in %s", line, path)
                    continue

                ts = None
                bytes_sent = None

                try:
                    ts = datetime.strptime(log["datetime"], "%d/%b/%Y:%H:%M:%S %z")
                    bytes_sent = log["bytes_sent"].strip("-") or 0
                except ValueError:
                    pass

                log.pop("datetime")
                log.pop("bytes_sent")

                yield WebserverAccessLogRecord(
                    ts=ts,
                    bytes_sent=bytes_sent,
                    **log,
                    source=path,
                    _target=self.target,
                )

    @export(record=WebserverErrorLogRecord)
    def error(self) -> Iterator[WebserverErrorLogRecord]:
        """Return contents of NGINX error log files in unified ``WebserverErrorLogRecord`` format.

        References:
            - https://nginx.org/en/docs/ngx_core_module.html#error_log
            - https://github.com/nginx/nginx/blob/master/src/core/ngx_log.c
        """
        target_tz = self.target.datetime.tzinfo

        for path in self.error_paths:
            path = path.resolve(strict=True)
            if not path.is_file():
                self.target.log.warning("File not found: %s", path)
                continue

            for line in open_decompress(path, "rt"):
                if not (line := line.strip()):
                    continue

                if not (match := RE_ERROR_LOG.search(line)):
                    self.target.log.warning("Unable to match NGINX error log message %r in %s", line, path)
                    continue

                log = match.groupdict()

                try:
                    ts = datetime.strptime(log["ts"], "%Y/%m/%d %H:%M:%S").replace(tzinfo=target_tz)
                except ValueError:
                    ts = None

                log.pop("ts")
                log.pop("tid")

                yield WebserverErrorLogRecord(
                    ts=ts,
                    **log,
                    source=path,
                    _target=self.target,
                )

    @export(record=WebserverHostRecord)
    def hosts(self) -> Iterator[WebserverHostRecord]:
        """Return found server directives in the NGINX configuration.

        References:
            - https://nginx.org/en/docs/http/ngx_http_core_module.html#server
        """

        def yield_record(current_server: dict) -> Iterator[WebserverHostRecord]:
            yield WebserverHostRecord(
                ts=host_path.lstat().st_mtime,
                server_name=current_server.get("server_name") or current_server.get("listen"),
                server_port=current_server.get("listen"),
                root_path=current_server.get("root"),
                access_log_config=current_server.get("access_log"),
                error_log_config=current_server.get("error_log"),
                source=host_path,
                _target=self.target,
            )

        for host_path in self.host_paths:
            current_server = {}
            seen_server_directive = False
            for line in host_path.open("rt"):
                if "server {" in line:
                    if current_server:
                        yield from yield_record(current_server)
                    current_server = {}
                    seen_server_directive = True

                elif seen_server_directive:
                    key, _, value = line.strip().partition(" ")
                    current_server[key] = value.rstrip(";")

            if current_server:
                yield from yield_record(current_server)


def parse_json_line(line: str) -> dict[str, str] | None:
    """Attempt to parse a default NGINX JSON log line.

    We assume the custom ``log_format`` uses the following default NGINX field names:

    .. code-block:: text

        time_local, time, remote_addr, remote_ip, remote_user, request_method, request,
        response, status, body_bytes_sent, request_time, http_referrer, referrer,
        http_user_agent, agent

    Unfortunately NGINX has no official default naming convention for JSON access logs,
    users can configure the JSON ``log_format`` as they see fit.

    References:
        - https://nginx.org/en/docs/http/ngx_http_log_module.html
        - https://github.com/elastic/examples/blob/master/Common%20Data%20Formats/nginx_json_logs/README.md
    """

    try:
        json_log = json.loads(line)
        return {
            "datetime": json_log.get("time_local") or json_log.get("time"),
            "remote_ip": json_log.get("remote_addr") or json_log.get("remote_ip"),
            "remote_user": json_log.get("remote_user"),
            "method": json_log.get("request_method"),
            "uri": json_log.get("request_uri") or json_log.get("request"),
            "status": json_log.get("status") or json_log.get("response"),
            "status_code": json_log.get("status") or json_log.get("response"),
            "referer": json_log.get("http_referrer") or json_log.get("referrer"),
            "useragent": json_log.get("http_user_agent") or json_log.get("agent"),
            "response_time_ms": json_log.get("request_time"),
            "bytes_sent": json_log.get("body_bytes_sent") or json_log.get("bytes"),
        }

    except json.JSONDecodeError as e:
        raise ValueError(f"Could not parse NGINX log line {line!r}: {e}") from e



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows import clfs
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_clfs_plugin(target_win: Target, fs_win: VirtualFilesystem) -> None:
    data_dir = absolute_path("_data/plugins/os/windows/clfs")

    fs_win.map_dir("windows/system32/config/", data_dir)

    target_win.add_plugin(clfs.ClfsPlugin)

    records = list(target_win.clfs())

    assert len(records) == 24

    expected_stream = "\\Device\\HarddiskVolume3\\wd\\compilerTemp\\BMT.SignCompDB.1lltmqvq.24r\\MetadataEsdGen\\mounted_image\\Windows\\System32\\config\\DRIVERS{53b39e70-18c4-11ea-a811-000d3aa4692b}.TM.blf"  # noqa: E501
    stream_names = {str(r.stream_name) for r in records}
    assert len(stream_names) == 1
    assert next(iter(stream_names)) == expected_stream

    stream_ids = {r.clfs_stream_id for r in records}
    assert len(stream_ids) == 1
    assert next(iter(stream_ids)) == 0

    expected_container = "DRIVERS{53b39e70-18c4-11ea-a811-000d3aa4692b}.TMContainer00000000000000000001.regtrans-ms"
    container_names = {str(r.container_name) for r in records}
    assert len(container_names) == 1
    assert next(iter(container_names)) == expected_container

    container_ids = {r.clfs_container_id for r in records}

    assert len(container_ids) == 1
    assert next(iter(container_ids)) == 0

    container_sizes = {r.container_size for r in records}
    assert len(container_sizes) == 1
    assert next(iter(container_sizes)) == 524288

    expected_first_record_offset = 36976
    first_record_offset = records[0].record_offset
    assert first_record_offset == expected_first_record_offset

    expected_last_record_offset = 3184
    last_record_offset = records[-1].record_offset
    assert last_record_offset == expected_last_record_offset

    expected_first_record_data = bytes.fromhex("000000000000000004010000762f16000519ea11a810000d3aa41ef300000000")
    first_record = records[0].record_data
    assert first_record == expected_first_record_data

    expected_first_block_data = bytes.fromhex(
        """
        04 01 00 00 04 00 00 00
        be 8e 6b c1 61 db ec 11
        a4 ba 00 50 56 ef c5 14
        0d 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        00 00 80 02 01 00 00 10
        00 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        """
    )
    first_block_data = records[0].block_data
    assert first_block_data == expected_first_block_data

    expected_last_record_data = bytes.fromhex("000000000000000004010000762f16000519ea11a810000d3aa41ef300000000")
    last_record = records[-1].record_data
    assert last_record == expected_last_record_data

    expected_last_block_data = bytes.fromhex(
        """
        04 01 00 00 04 00 00 00
        e1 b9 29 a6 ea d7 eb 11
        a4 6b 3c 22 fb 13 6b f1
        21 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        00 00 80 02 01 00 00 10
        00 00 00 00 00 00 00 00
        00 00 00 00 00 00 00 00
        """
    )
    last_block_data = records[-1].block_data
    assert last_block_data == expected_last_block_data



from __future__ import annotations

import argparse
import os
import pathlib
import platform
import re
import sys
from collections import ChainMap
from io import BytesIO, StringIO
from pathlib import Path
from typing import IO, TYPE_CHECKING, Callable, TextIO
from unittest.mock import MagicMock, call, mock_open, patch

import pytest

from dissect.target.helpers.fsutil import TargetPath, normalize
from dissect.target.tools import fsutils
from dissect.target.tools.shell import (
    TargetCli,
    TargetHubCli,
    build_pipe,
    build_pipe_stdout,
)
from dissect.target.tools.shell import main as target_shell
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

try:
    import pexpect
    import pexpect.expect

    HAS_PEXPECT = True
except ImportError:
    HAS_PEXPECT = False

GREP_MATCH = "test1 and test2"
GREP_MISSING = "test2 alone"
INPUT_STREAM = f"""
    line 1
    test1 alone
    {GREP_MATCH}
    {GREP_MISSING}
    last line
"""


@pytest.mark.skipif(platform.system() == "Windows", reason="Unix-specific test.")
def test_build_pipe() -> None:
    pipeparts = [
        "grep",
        "test2",
        "|",
        "grep",
        "test1",
    ]

    with build_pipe(pipeparts) as (pipe_stdin, pipe_stdout):
        print(INPUT_STREAM, file=pipe_stdin)

    output = pipe_stdout.read().decode("utf-8").strip()
    assert output == GREP_MATCH


@pytest.mark.skipif(platform.system() == "Windows", reason="Unix-specific test.")
def test_build_pipe_nonexisting_command() -> None:
    dummy_command = "non-existing-command"
    pipeparts = ["grep", "test1", "|", dummy_command]
    input_stream = "input data test1"

    with (
        pytest.raises(OSError, match="No such file or directory: 'non-existing-command'"),
        build_pipe_stdout(pipeparts) as pipe_stdin,
    ):
        print(input_stream, file=pipe_stdin)

    with (
        pytest.raises(OSError, match="No such file or directory: 'non-existing-command'"),
        build_pipe(pipeparts) as (pipe_stdin, _),
    ):
        print(input_stream, file=pipe_stdin)


@pytest.mark.skipif(platform.system() == "Windows", reason="Unix-specific test.")
def test_build_pipe_broken_pipe() -> None:
    # `ls` command does not accept stdin, so pipe closes prematurely
    pipeparts = ["grep", "test1", "|", "ls"]
    input_stream = "input data test1"

    # there should be no errors raised
    with build_pipe(pipeparts) as (pipe_stdin, _):
        print(input_stream, file=pipe_stdin)

    # there should be no errors raised
    with build_pipe_stdout(pipeparts) as pipe_stdin:
        print(input_stream, file=pipe_stdin)


def test_targethubcli_autocomplete_enter(make_mock_targets: Callable) -> None:
    target1, target2 = make_mock_targets(2)

    target1.hostname = "dev-null-1.localhost"
    target2.hostname = "dev-null-2.localhost"

    hub_cli = TargetHubCli([target1, target2], TargetCli)
    suggestions = hub_cli.complete_enter("dev-", "enter dev-", 6, 10)
    assert suggestions == [
        "dev-null-1.localhost",
        "dev-null-2.localhost",
    ]

    suggestions = hub_cli.complete_enter("dev-null-1", "enter dev-null-1", 6, 16)
    assert suggestions == ["dev-null-1.localhost"]

    suggestions = hub_cli.complete_enter("xxx-", "enter xxx-", 6, 10)
    assert suggestions == []

    suggestions = hub_cli.complete_enter("1", "enter 1", 6, 7)
    assert suggestions == ["1"]


def test_targetcli_autocomplete(target_bare: Target, monkeypatch: pytest.MonkeyPatch) -> None:
    target_cli = TargetCli(target_bare)

    mock_subfolder = MagicMock(spec_set=TargetPath)
    mock_subfolder.is_dir.return_value = True
    mock_subfile = MagicMock(spec_set=TargetPath)
    mock_subfile.is_dir.return_value = False

    base_path = "/base-path"

    subfolder_name = "subfolder"
    subfile_name = "subfile"
    subpath_mismatch = "mismatch"

    def dummy_scandir(path: TargetPath) -> list[tuple[TargetPath | None, str]]:
        assert str(path) == base_path
        return [
            (mock_subfolder, subfolder_name),
            (mock_subfile, subfile_name),
            (None, subpath_mismatch),
        ]

    monkeypatch.setattr("dissect.target.tools.shell.ls_scandir", dummy_scandir)
    suggestions = target_cli.completedefault("sub", f"ls {base_path}/sub", 3 + len(base_path), 3 + len(base_path) + 3)

    # We expect folder suggestions to be trailed with a '/'
    assert suggestions == [f"{subfolder_name}/", subfile_name]


@pytest.fixture
def targetrc_file() -> Iterator[list[str]]:
    content = """
    ls
        # This is a comment line and should be ignored
    ll
    """

    original_open = pathlib.Path.open

    def custom_open(self: Path, *args, **kwargs) -> IO:
        if self.name.endswith(".targetrc"):
            return mock_open(read_data=content)()
        return original_open(self, *args, **kwargs)

    with patch("pathlib.Path.open", custom_open):
        yield ["ls", "ll"]


def test_targetcli_targetrc(target_bare: Target, targetrc_file: list[str]) -> None:
    with patch.object(TargetCli, "onecmd", autospec=True) as mock_onecmd:
        cli = TargetCli(target_bare)

        cli.preloop()

        expected_calls = [call(cli, cmd) for cmd in targetrc_file]
        mock_onecmd.assert_has_calls(expected_calls, any_order=False)


@pytest.mark.skipif(platform.system() == "Windows", reason="Unix-specific test.")
def test_pipe_symbol_parsing(capfd: pytest.CaptureFixture, target_bare: Target) -> None:
    cli = TargetCli(target_bare)

    def mock_func(func_args: list[str], func_stdout: TextIO) -> None:
        # if the operation of parsing out the first `|` was successful,
        # `mock_func` should receive an empty `func_args` list
        assert len(func_args) == 0

        print(INPUT_STREAM, file=func_stdout)

    command_args_str = "|grep test2|grep test1"

    cli._exec(mock_func, command_args_str)

    sys.stdout.flush()
    sys.stderr.flush()

    captured = capfd.readouterr()

    assert GREP_MATCH in captured.out
    assert GREP_MISSING not in captured.out


@pytest.mark.skipif(platform.system() == "Windows", reason="Unix-specific test.")
def test_exec_target_command(capfd: pytest.CaptureFixture, target_default: Target) -> None:
    cli = TargetCli(target_default)
    # `users` from the general OSPlugin does not ouput any records, but as the
    # ouput is piped to, the records are transformed to a binary record stream,
    # so we pipe it through rdump to get a correct count of 0 from wc.
    cli.default("users | rdump | wc -l")

    sys.stdout.flush()
    sys.stderr.flush()

    captured = capfd.readouterr()

    assert captured.out.endswith("0\n")


def test_target_cli_ls(target_win: Target, capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    # disable colorful output in `target-shell`
    monkeypatch.setattr(fsutils, "LS_COLORS", {})

    cli = TargetCli(target_win)
    cli.onecmd("ls")

    captured = capsys.readouterr()
    assert captured.out == "c:\nsysvol" + "\n"


@pytest.mark.parametrize(
    ("folders", "files", "save", "expected"),
    [
        ("/a/b/c|/d", "/a/1.txt|/a/b/2.txt|/d/3.txt", "/a", "a|a/1.txt|a/b|a/b/2.txt|a/b/c"),
        ("/a/b/c|/d", "/a/1.txt|/b/2.txt|/d/3.txt", "/d", "d|d/3.txt"),
        ("/a/b/c|/d", "/a/1.txt|/b/2.txt|/d/3.txt", "/b", "b|b/2.txt"),
        ("/p/q/n", "/p/q/n/1.txt", "/p/q/n/1.txt", "1.txt"),
        ("/p/q/n", "/p/q/n/1.txt|2.txt", "/p/q/n/1.txt", "1.txt"),
        (
            "/save_test/bin|save_test/data",
            "/save_test/data/hello|/save_test/bin/world",
            "/save_test",
            "save_test|save_test/bin|save_test/bin/world|save_test/data|save_test/data/hello",
        ),
    ],
)
def test_target_cli_save(
    target_win: Target, tmp_path: Path, folders: str, files: str, save: str, expected: str
) -> None:
    output_dir = tmp_path / "output"
    output_dir.mkdir(parents=True, exist_ok=True)

    cli = TargetCli(target_win)
    for folder in folders.split("|"):
        target_win.fs.root.makedirs(folder)
    for _file in files.split("|"):
        target_win.fs.root.map_file_fh(_file, BytesIO(_file.encode("utf-8")))
    args = argparse.Namespace(path=[save], out=output_dir, verbose=False)
    cli.cmd_save(args, sys.stdout)

    def _map_function(path: Path) -> str:
        relative_path = str(path.relative_to(output_dir))
        return normalize(relative_path, alt_separator=target_win.fs.alt_separator)

    path_map = (_map_function(path) for path in output_dir.rglob("*"))
    tree = "|".join(sorted(path_map))

    assert tree == expected


def run_target_shell(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture, argv: str | list, stdin: str
) -> tuple[bytes, bytes]:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-shell"] + (argv if isinstance(argv, list) else [argv]))
        m.setattr("sys.stdin", StringIO(stdin))
        m.setenv("NO_COLOR", "1")
        target_shell()
        return capsys.readouterr()


@pytest.mark.parametrize(
    ("provided_input", "expected_output"),
    [
        ("hello", "world.txt"),  # Latin
        ("ħēļľŏ", "ŵőřŀđ.txt"),  # Latin Extended-A
        ("مرحبًا", "عالم.txt"),  # Arabic
        ("你好", "世界.txt"),  # Chinese Simplified
        ("привет", "мир.txt"),  # Cyrillic
        ("🕵🕵🕵", "👀👀👀.txt"),  # Emoji
    ],
)
def test_target_cli_unicode_argparse(
    capsys: pytest.CaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    provided_input: str,
    expected_output: str,
) -> None:
    out, err = run_target_shell(
        monkeypatch,
        capsys,
        str(absolute_path("_data/tools/shell/unicode.tar")),
        f"ls unicode/charsets/{provided_input}",
    )
    out = out.replace("unicode.tar:/$", "").strip()
    assert out == expected_output
    assert "unrecognized arguments" not in err


def test_shell_cmd_alias(monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture) -> None:
    """Test if alias commands call their parent attribute correctly."""
    target_path = str(absolute_path("_data/tools/info/image.tar"))

    # 'dir' and 'ls' should return the same output
    dir_out, _ = run_target_shell(monkeypatch, capsys, target_path, "dir")
    ls_out, _ = run_target_shell(monkeypatch, capsys, target_path, "ls")
    assert dir_out == ls_out

    # ll is not really a standard aliased command so we test that separately.
    ls_la_out, _ = run_target_shell(monkeypatch, capsys, target_path, "ls -la")
    ll_out, _ = run_target_shell(monkeypatch, capsys, target_path, "ll")
    assert ls_la_out == ll_out


def test_shell_cli_command(monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture) -> None:
    target_path = str(absolute_path("_data/tools/info/image.tar"))
    dir_out, _ = run_target_shell(monkeypatch, capsys, target_path, "dir")
    ls_out, _ = run_target_shell(monkeypatch, capsys, [target_path, "-c", "dir"], "")
    assert dir_out == "ubuntu:/$ " + ls_out + "ubuntu:/$ \n"


def test_shell_cmd_alias_runtime(monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture) -> None:
    """Test if alias commands call their parent attribute correctly."""
    target_path = str(absolute_path("_data/tools/info/image.tar"))

    # 'list' and 'ls' should return the same output after runtime aliasing
    list_out, _ = run_target_shell(monkeypatch, capsys, target_path, "alias list=ls xxl='ls -la'\nlist")
    sys.stdout.flush()
    ls_out, _ = run_target_shell(monkeypatch, capsys, target_path, "ls")
    assert list_out == "ubuntu:/$ " + ls_out

    # list aliases
    sys.stdout.flush()
    out, _ = run_target_shell(monkeypatch, capsys, target_path, "alias")
    assert out == "ubuntu:/$ alias list=ls\nalias xxl=ls -la\nubuntu:/$ \n"

    # list single aliases
    sys.stdout.flush()
    out, _ = run_target_shell(monkeypatch, capsys, target_path, "alias list")
    assert out == "ubuntu:/$ alias list=ls\nubuntu:/$ \n"

    # unalias
    sys.stdout.flush()
    run_target_shell(monkeypatch, capsys, target_path, "unalias xxl")
    out, _ = run_target_shell(monkeypatch, capsys, target_path, "alias")
    assert out == "ubuntu:/$ alias list=ls\nubuntu:/$ \n"

    # unalias multiple and non-existant
    sys.stdout.flush()
    out, _ = run_target_shell(monkeypatch, capsys, target_path, "unalias list abc")
    assert out == "ubuntu:/$ alias abc not found\nubuntu:/$ \n"

    # alias multiple broken - b will be empty
    sys.stdout.flush()
    run_target_shell(monkeypatch, capsys, target_path, "alias a=1 b=")
    out, _ = run_target_shell(monkeypatch, capsys, target_path, "alias")
    assert out == "ubuntu:/$ alias a=1\nalias b=\nubuntu:/$ \n"

    # alias set/get mixed
    sys.stdout.flush()
    out, _ = run_target_shell(monkeypatch, capsys, target_path, "alias b=1 a")
    assert out == "ubuntu:/$ alias a=1\nubuntu:/$ \n"

    # alias with other symbols not allowed due to parser difference
    sys.stdout.flush()
    out, _ = run_target_shell(monkeypatch, capsys, target_path, "alias b+1")
    assert out.find("*** Unhandled error: Token not allowed") > -1


def test_shell_hostname_escaping(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture, tmp_path: Path
) -> None:
    """Test if we properly escape hostnames in the base prompt."""

    tmp_path.joinpath("etc").mkdir()
    tmp_path.joinpath("var").mkdir()
    tmp_path.joinpath("opt").mkdir()
    tmp_path.joinpath("etc/hostname").write_bytes(b"hostname\x00\x01\x02\x03")

    sys.stdout.flush()
    out, _ = run_target_shell(monkeypatch, capsys, str(tmp_path), "\n")

    assert "hostname\\x00\\x01\\x02\\x03" in out


@pytest.mark.skipif(not HAS_PEXPECT, reason="requires pexpect")
@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="pexpect.spawn not available on Windows",
)
def test_shell_prompt_tab_autocomplete() -> None:
    """Test the prompt tab-autocompletion."""

    ANSI_ESCAPE = re.compile(rb"\x07|\x08|\x0d|\x7f|\x1b[@-_][0-?]*[ -/]*[@-~]")
    original_new_data = pexpect.expect.Expecter.new_data

    def ansi_new_data(cls: pexpect.expect.Expecter, data: bytes) -> int | None:
        """Matches on carriage returns (``\x0d`` / ``\r``) so change your ``child.expect()`` calls accordingly."""
        return original_new_data(cls, ANSI_ESCAPE.sub(b"", data))

    target_path = absolute_path("_data/tools/info/image.tar")

    with patch("pexpect.expect.Expecter.new_data", new=ansi_new_data):
        # We set NO_COLOR=1 so that the output is not colored and easier to match
        child = pexpect.spawn("target-shell", args=[str(target_path)], env=ChainMap(os.environ, {"NO_COLOR": "1"}))

        # increase window size to avoid line wrapping
        child.setwinsize(100, 100)

        if platform.python_implementation() == "PyPy":
            major, minor, _patch = tuple(map(int, platform.python_version_tuple()))
            if major < 3 or (major == 3 and (minor < 10 or (minor == 10 and _patch < 14))):
                child.expect_exact(
                    "Note for users of PyPy < 3.10.14:\n"
                    "Autocomplete might not work due to an outdated version of pyrepl/readline.py\n"
                    "To fix this, please update your version of PyPy.\n",
                    timeout=30,
                )
                child.kill(9)  # 🔫
                return

            pytest.skip("PyPy in CI does not have a functional readline")

        child.expect_exact("ubuntu:/$ ", timeout=30)
        # this should auto complete to `ls /home/user`
        child.send("ls /home/u\t")
        # expect the prompt to be printed again
        child.expect_exact("ls /home/user/", timeout=5)
        # execute the autocompleted command
        child.send("\n")
        # we expect the files in /home/user to be printed
        child.expect_exact(".bash_history\n.zsh_history\n", timeout=5)
        child.expect_exact("ubuntu:/$ ", timeout=5)

        # send partial ls /etc/ command
        child.send("ls /etc/")

        # we send two TABS to get the list of files in /etc/
        child.send("\t\t")

        # expect the files in /etc/ to be printed
        child.expect(
            r"hosts\s+localtime\s+network\/\s+os-release\s+passwd\s+shadow\s+timezone\s*ubuntu:\/\$ ls \/etc\/",
            timeout=5,
        )

        # send newline to just list everything in /etc/
        child.send("\n")
        # expect the last few files in /etc/ to be printed
        child.expect_exact("shadow\ntimezone\n", timeout=5)

        # exit the shell
        child.expect_exact("ubuntu:/$ ", timeout=5)
        child.sendline("exit")
        child.expect(pexpect.EOF, timeout=5)



#!/usr/bin/env python
from __future__ import annotations

import argparse
import json
import logging
from datetime import datetime
from pathlib import Path

from dissect.target.exceptions import TargetError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.target import Target
from dissect.target.tools.query import record_output
from dissect.target.tools.utils import (
    catch_sigpipe,
    configure_generic_arguments,
    process_generic_arguments,
)

InfoRecord = TargetRecordDescriptor(
    "target/info",
    [
        ("datetime", "last_activity"),
        ("datetime", "install_date"),
        ("net.ipaddress[]", "ips"),
        ("string", "os_family"),
        ("string", "os_version"),
        ("string", "architecture"),
        ("string[]", "language"),
        ("string", "timezone"),
        ("string[]", "disks"),
        ("string[]", "volumes"),
        ("string[]", "mounts"),
        ("string[]", "children"),
    ],
)


log = logging.getLogger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="target-info",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("targets", metavar="TARGETS", nargs="*", help="Targets to display info from")
    parser.add_argument("--children", action="store_true", help="include children")
    parser.add_argument("--from-file", nargs="?", type=Path, help="file containing targets to load")
    parser.add_argument("-s", "--strings", action="store_true", help="print output as string")
    parser.add_argument("-r", "--record", action="store_true", help="print output as record")
    parser.add_argument("-j", "--json", action="store_true", help="output records as pretty json")
    parser.add_argument("-J", "--jsonlines", action="store_true", help="output records as one-line json")
    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()

    if not args.targets and not args.from_file:
        parser.error("too few arguments")

    if args.from_file:
        if not args.from_file.exists():
            parser.error(f"--from-file {args.from_file} does not exist")

        targets = args.from_file.read_text().splitlines()
        while targets[-1] == "":
            targets = targets[:-1]
        args.targets = targets

    process_generic_arguments(args)

    try:
        for i, target in enumerate(Target.open_all(args.targets, include_children=args.children)):
            try:
                if args.jsonlines:
                    print(json.dumps(get_target_info(target), default=str))
                elif args.json:
                    print(json.dumps(get_target_info(target), indent=4, default=str))
                elif args.record:
                    rs = record_output(args.strings)
                    rs.write(InfoRecord(**get_target_info(target), _target=target))
                else:
                    if i > 0:
                        print("-" * 70)
                    print_target_info(target)
            except Exception as e:  # noqa: PERF203
                target.log.error("Exception in retrieving information for target: `%s`, use `-vv` for details", target)  # noqa: TRY400
                target.log.debug("", exc_info=e)
    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    return 0


def get_target_info(target: Target) -> dict[str, str | list[str]]:
    return {
        "disks": get_disks_info(target),
        "volumes": get_volumes_info(target),
        "mounts": get_mounts_info(target),
        "children": get_children_info(target),
        "hostname": target.hostname,
        "domain": get_optional_func(target, "domain"),
        "ips": target.ips,
        "os_family": target.os,
        "os_version": target.version,
        "architecture": target.architecture,
        "language": get_optional_func(target, "language"),
        "timezone": get_optional_func(target, "timezone"),
        "install_date": get_optional_func(target, "install_date"),
        "last_activity": get_optional_func(target, "activity"),
    }


def get_optional_func(target: Target, func: str) -> str | None:
    if target.has_function(func):
        return getattr(target, func)
    return None


def print_target_info(target: Target) -> None:
    print(target)

    for name, value in get_target_info(target).items():
        if name in ["disks", "volumes", "mounts", "children"]:
            if not any(value):
                continue
            print(f"\n{name.capitalize()}")
            for i in value:
                values = " ".join([f"{k}={v!r}" for k, v in i.items()])
                print(f"- <{name.capitalize()[:-1].replace('re', '')} {values}>")
            continue

        if isinstance(value, list):
            value = ", ".join(map(str, value))

        if isinstance(value, datetime):
            value = value.isoformat(timespec="microseconds")

        if name == "hostname":
            print()

        print(f"{name.capitalize().replace('_', ' '):14s} : {value}")


def get_disks_info(target: Target) -> list[dict[str, str | int]]:
    return [{"type": d.__type__, "size": d.size} for d in target.disks]


def get_volumes_info(target: Target) -> list[dict[str, str | int | None]]:
    return [{"name": v.name, "size": v.size, "fs": v.fs.__type__ if v.fs else None} for v in target.volumes]


def get_mounts_info(target: Target) -> list[dict[str, str | None]]:
    return [{"fs": fs.__type__, "path": path} for path, fs in target.fs.mounts.items()]


def get_children_info(target: Target) -> list[dict[str, str]]:
    return [{"type": c.type, "path": str(c.path)} for c in target.list_children()]


if __name__ == "__main__":
    main()



from __future__ import annotations

import io
from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.child.wsl import WSLChildTargetPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_wsl(target_win_users: Target, hive_hku: VirtualHive, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file_fh(
        "users/john/appdata/local/Packages/CanonicalGroupLimited.Ubuntu22.04LTS_79rhkp1fndgsc/LocalState/ext4.vhdx",
        io.BytesIO(),
    )

    wsl_keys_name = "Software\\Microsoft\\Windows\\CurrentVersion\\Lxss"
    wsl_keys = VirtualKey(hive_hku, wsl_keys_name)

    wsl_key = VirtualKey(hive_hku, "{12345678-1234-1234-1234-123456789012}")
    wsl_key.add_value(
        "BasePath",
        VirtualValue(
            hive_hku,
            "BasePath",
            "C:\\Users\\John\\AppData\\Local\\Packages\\CanonicalGroupLimited.Ubuntu22.04LTS_79rhkp1fndgsc\\LocalState",
        ),
    )

    wsl_keys.add_subkey(wsl_key.name, wsl_key)
    hive_hku.map_key(wsl_keys_name, wsl_keys)

    target_win_users.add_plugin(WSLChildTargetPlugin)

    children = list(target_win_users.list_children())
    assert len(children) == 1
    assert children[0].type == "wsl"
    assert (
        str(children[0].path)
        == "C:\\Users\\John\\AppData\\Local\\Packages\\CanonicalGroupLimited.Ubuntu22.04LTS_79rhkp1fndgsc\\LocalState\\ext4.vhdx"  # noqa E501
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.evidence import AsdfSnapshot
from dissect.evidence.asdf.asdf import IDX_METADATA

from dissect.target.containers.asdf import AsdfContainer
from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class AsdfLoader(Loader):
    """Load an ASDF target."""

    METADATA_PREFIX = "$asdf$"

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.asdf = AsdfSnapshot(path.open("rb"))

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".asdf"

    def map(self, target: Target) -> None:
        for disk in self.asdf.disks():
            target.disks.add(AsdfContainer(disk))

        target.fs.mount(self.METADATA_PREFIX, TarFilesystem(self.asdf.open(IDX_METADATA)))



from __future__ import annotations

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.plugin import Plugin, export


class ExchangePlugin(Plugin):
    """Microsoft Exchange Server plugin."""

    __namespace__ = "exchange"

    def check_compatible(self) -> None:
        if not len(self.install_paths()):
            raise UnsupportedPluginError("No Exchange install path found")

    def install_paths(self) -> list[str]:
        paths = []
        key = "HKLM\\SOFTWARE\\Microsoft\\ExchangeServer"
        for reg_key in self.target.registry.keys(key):
            for subkey in reg_key.subkeys():
                try:
                    setup_key = subkey.subkey("Setup")
                    install_path = setup_key.value("MsiInstallPath").value
                    paths.append(install_path)
                except RegistryError:  # noqa: PERF203
                    pass

        return paths

    @export(output="none")
    def transport_agents(self) -> None:
        """Print the content of the config file for Transport Agents for Microsoft Exchange.

        A Transport Agent is additional software on a Microsoft Exchange server that allows for custom processing of
        email messages that go through the transport pipeline.

        References:
            - https://docs.microsoft.com/en-us/exchange/mail-flow/transport-agents/transport-agents?view=exchserver-2019
        """
        for path in self.install_paths():
            config_path = self.target.fs.path(path).joinpath("TransportRoles/Agents/agents.config")
            if not config_path.exists():
                continue

            print(config_path.open().read())



from __future__ import annotations

import importlib.util
import pathlib
import tempfile
import textwrap
from io import BytesIO
from itertools import chain
from typing import TYPE_CHECKING, Callable

import pytest

from dissect.target import container, filesystem, loader, plugin, volume
from dissect.target.exceptions import RegistryKeyNotFoundError
from dissect.target.filesystem import Filesystem, VirtualFilesystem, VirtualSymlink
from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.helpers import keychain
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugin import _generate_long_paths, _os_match
from dissect.target.plugins.os.default._os import DefaultOSPlugin
from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.plugins.os.unix.bsd.citrix._os import CitrixPlugin
from dissect.target.plugins.os.unix.bsd.darwin.ios._os import IOSPlugin
from dissect.target.plugins.os.unix.bsd.darwin.macos._os import MacOSPlugin
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.plugins.os.unix.linux.android._os import AndroidPlugin
from dissect.target.plugins.os.unix.linux.debian._os import DebianPlugin
from dissect.target.plugins.os.unix.linux.redhat._os import RedHatPlugin
from dissect.target.plugins.os.unix.linux.suse._os import SuSEPlugin
from dissect.target.plugins.os.windows import registry
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.plugin import OSPlugin


HAS_BENCHMARK = importlib.util.find_spec("pytest_benchmark") is not None


def pytest_configure(config: pytest.Config) -> None:
    if not HAS_BENCHMARK:
        # If we don't have pytest-benchmark (or pytest-codspeed) installed, register the benchmark marker ourselves
        # to avoid pytest warnings
        config.addinivalue_line("markers", "benchmark: mark test for benchmarking (requires pytest-benchmark)")


def pytest_sessionstart(session: pytest.Session) -> None:
    # Test if the _data/ directory is present and if not, as is the case in Python
    # source distributions of dissect.target, we give an error
    data_dir = absolute_path("_data")
    if not data_dir.is_dir():
        session.shouldfail = (
            "! !\n"
            f"No test data directory {data_dir} found.\n"
            "This can happen when you have downloaded the source distribution\n"
            "of dissect.target from pypi.org. If so, retrieve the test data from\n"
            "the dissect.target GitHub repository at:\n"
            "https://github.com/fox-it/dissect.target"
            "\n! !"
        )
    else:
        # Test if the test data looks like LFS references and if so, we give an error
        # This can happen when git-lfs is not installed or not configured correctly
        for file in (path for path in data_dir.rglob("*") if path.is_file()):
            with file.open("rb") as fh:
                if fh.read(42) == b"version https://git-lfs.github.com/spec/v1":
                    session.shouldfail = (
                        "! !\n"
                        "Test data files look like git-lfs references.\n"
                        "This can happen when git-lfs is not installed or not configured correctly.\n"
                        "Install git-lfs and run: \n"
                        "git lfs install && git lfs pull"
                        "\n! !"
                    )
            break


def pytest_runtest_setup(item: pytest.Item) -> None:
    if not HAS_BENCHMARK and item.get_closest_marker("benchmark") is not None:
        pytest.skip("pytest-benchmark is not installed")


@pytest.fixture(autouse=True)
def clear_caches() -> None:
    _generate_long_paths.cache_clear()
    _os_match.cache_clear()


@pytest.fixture(autouse=True)
def clear_lazy_imports() -> None:
    """Clear lazy imports before each test to ensure a clean state."""

    for lazy_attr in chain(
        loader.LOADERS_BY_SCHEME.values(),
        filesystem.FILESYSTEMS,
        container.CONTAINERS,
        volume.LOGICAL_VOLUME_MANAGERS,
        volume.ENCRYPTED_VOLUME_MANAGERS,
    ):
        lazy_attr._loaded = False
        lazy_attr._exc = None
        lazy_attr._module._module = None
        lazy_attr._module._loaded = False


CLEAN_PLUGINS = plugin.generate()


@pytest.fixture(autouse=True)
def reset_plugins() -> None:
    plugin.PLUGINS = CLEAN_PLUGINS


def make_mock_target(tmp_path: pathlib.Path) -> Iterator[Target]:
    with tempfile.NamedTemporaryFile(dir=tmp_path, prefix="MockTarget-", delete=False) as tmp_file:
        tmp_file.close()
        target = Target()
        target.path = pathlib.Path(tmp_file.name)
        yield target


def make_os_target(
    tmp_path: pathlib.Path,
    os_plugin: type[OSPlugin],
    root_fs: Filesystem | None = None,
    apply_target: bool = True,
) -> Target:
    mock_target = next(make_mock_target(tmp_path))
    mock_target._os_plugin = os_plugin

    if root_fs is not None:
        mock_target.filesystems.add(root_fs)

    if apply_target:
        mock_target.apply()

    return mock_target


@pytest.fixture
def make_mock_targets(request: pytest.FixtureRequest, tmp_path: pathlib.Path) -> Callable[[int], Iterator[Target]]:
    def _make_targets(size: int) -> Iterator[Target]:
        for _ in range(size):
            yield from make_mock_target(tmp_path)

    return _make_targets


@pytest.fixture
def fs_win(tmp_path: pathlib.Path) -> VirtualFilesystem:
    fs = VirtualFilesystem(case_sensitive=False, alt_separator="\\")
    fs.map_dir("windows/system32", tmp_path)
    fs.map_dir("windows/system32/config/", tmp_path)
    return fs


@pytest.fixture
def fs_unix() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc")
    return fs


@pytest.fixture
def fs_linux() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc")
    fs.makedirs("opt")
    return fs


@pytest.fixture
def fs_debian() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc/dpkg")
    fs.makedirs("opt")
    return fs


@pytest.fixture
def fs_redhat() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc/sysconfig/network-scripts")
    fs.makedirs("opt")
    return fs


@pytest.fixture
def fs_suse() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc/zypp")
    fs.makedirs("opt")
    return fs


@pytest.fixture
def fs_linux_sys(fs_linux: VirtualFilesystem) -> VirtualFilesystem:
    fs_linux.makedirs("sys")
    return fs_linux


@pytest.fixture
def fs_linux_proc(fs_linux: VirtualFilesystem) -> VirtualFilesystem:
    fs = fs_linux

    procs = (
        (
            "proc/1",
            VirtualSymlink(fs, "/proc/1/fd/4", "socket:[1337]"),
            "test\x00cmdline\x00",
            "VAR=1",
        ),
        (
            "proc/2",
            VirtualSymlink(fs, "/proc/2/fd/4", "socket:[1338]"),
            "\x00",
            "VAR=1\x00",
        ),
        (
            "proc/3",
            VirtualSymlink(fs, "/proc/3/fd/4", "socket:[1339]"),
            "sshd",
            "VAR=1",
        ),
        (
            "proc/1337",
            VirtualSymlink(fs, "/proc/1337/fd/4", "socket:[1337]"),
            "acquire\x00-p\x00full\x00--proc\x00",
            "VAR=1",
        ),
    )
    stat_files_data = (
        "1 (systemd) S 0 1 1 0 -1 4194560 53787 457726 166 4255 112 260 761 548 20 0 1 0 30 184877056 2658 18446744073709551615 93937510957056 93937511789581 140726499200496 0 0 0 671173123 4096 1260 0 0 0 17 0 0 0 11 0 0 93937512175824 93937512476912 93937519890432 140726499204941 140726499204952 140726499204952 140726499205101 0\n",  # noqa
        "2 (kthread) K 1 1 1 0 -1 4194560 53787 457726 166 4255 112 260 761 548 20 0 1 0 30 184877056 2658 18446744073709551615 93937510957056 93937511789581 140726499200496 0 0 0 671173123 4096 1260 0 0 0 17 0 0 0 11 0 0 93937512175824 93937512476912 93937519890432 140726499204941 140726499204952 140726499204952 140726499205101 0\n",  # noqa
        "3 (sshd) W 1 2 1 0 -1 4194560 53787 457726 166 4255 112 260 761 548 20 0 1 0 30 184877056 2658 18446744073709551615 93937510957056 93937511789581 140726499200496 0 0 0 671173123 4096 1260 0 0 0 17 0 0 0 11 0 0 93937512175824 93937512476912 93937519890432 140726499204941 140726499204952 140726499204952 140726499205101 0\n",  # noqa
        "1337 (acquire) R 3 1 1 0 -1 4194560 53787 457726 166 4255 112 260 761 548 20 0 1 0 30 184877056 2658 18446744073709551615 93937510957056 93937511789581 140726499200496 0 0 0 671173123 4096 1260 0 0 0 17 0 0 0 11 0 0 93937512175824 93937512476912 93937519890432 140726499204941 140726499204952 140726499204952 140726499205101 0\n",  # noqa
    )

    for idx, proc in enumerate(procs):
        dir, fd, cmdline, environ = proc
        fs.makedirs(dir)
        fs.map_file_entry(fd.path, fd)

        fs.map_file_fh(dir + "/stat", BytesIO(stat_files_data[idx].encode()))
        fs.map_file_fh(dir + "/cmdline", BytesIO(cmdline.encode()))
        fs.map_file_fh(dir + "/environ", BytesIO(environ.encode()))

    # symlink acquire process to self
    fs.link("/proc/1337", "/proc/self")

    # boottime and uptime are needed for for time tests
    fs.map_file_fh("/proc/uptime", BytesIO(b"134368.27 132695.52\n"))
    fs.map_file_fh("/proc/stat", BytesIO(b"btime 1680559854"))

    return fs


@pytest.fixture
def fs_linux_proc_sockets(fs_linux_proc: VirtualFilesystem) -> VirtualFilesystem:
    fs = fs_linux_proc

    for filename in ("unix", "packet", "raw6", "raw", "udp6", "udp", "tcp6", "tcp"):
        fs.map_file(f"/proc/net/{filename}", absolute_path(f"_data/plugins/os/unix/linux/proc/net/{filename}"))

    return fs


@pytest.fixture
def fs_macos() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.makedirs("Applications")
    fs.makedirs("Library")
    return fs


@pytest.fixture
def fs_ios() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.makedirs("/private/var/preferences")
    fs.makedirs("/private/var/mobile")
    return fs


@pytest.fixture
def fs_bsd() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.map_file("/bin/freebsd-version", absolute_path("_data/plugins/os/unix/bsd/freebsd/freebsd-freebsd-version"))
    return fs


@pytest.fixture
def fs_android() -> VirtualFilesystem:
    fs = VirtualFilesystem()
    fs.makedirs("/data")
    fs.makedirs("/system")
    fs.makedirs("/vendor")
    fs.makedirs("/product")
    fs.map_file("/build.prop", absolute_path("_data/plugins/os/unix/linux/android/build.prop"))
    return fs


@pytest.fixture
def hive_hklm() -> VirtualHive:
    hive = VirtualHive()

    # set current control set to ControlSet001 and mock it
    change_controlset(hive, 1)

    # set windir
    current_version_key = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
    hive.map_key(current_version_key, VirtualKey(hive, current_version_key))
    hive.map_value(current_version_key, "SystemRoot", VirtualValue(hive, "SystemRoot", "c:\\Windows"))

    return hive


def change_controlset(hive: VirtualHive, num: int) -> None:
    """Update the current control set of the given HKLM hive."""

    if not isinstance(num, int) or num > 999 or num < 1:
        raise ValueError("ControlSet integer must be between 1 and 999")

    controlset_key = f"SYSTEM\\ControlSet{num:>03}"
    hive.map_key(controlset_key, VirtualKey(hive, controlset_key))

    select_key = "SYSTEM\\Select"
    hive.map_key(select_key, VirtualKey(hive, select_key))
    hive.map_value(select_key, "Current", VirtualValue(hive, "Current", num))


@pytest.fixture
def hive_hku() -> VirtualHive:
    return VirtualHive()


@pytest.fixture
def target_bare(tmp_path: pathlib.Path) -> Iterator[Target]:
    # A target without any associated OSPlugin
    yield from make_mock_target(tmp_path)


@pytest.fixture
def target_default(tmp_path: pathlib.Path) -> Target:
    return make_os_target(tmp_path, DefaultOSPlugin)


@pytest.fixture
def target_win(tmp_path: pathlib.Path, hive_hklm: VirtualHive, fs_win: Filesystem) -> Target:
    mock_target = make_os_target(tmp_path, WindowsPlugin, root_fs=fs_win, apply_target=False)

    mock_target.add_plugin(registry.RegistryPlugin, check_compatible=False)
    mock_target.registry.add_hive(
        "HKEY_LOCAL_MACHINE",
        "HKEY_LOCAL_MACHINE",
        hive_hklm,
        TargetPath(mock_target.fs, ""),
    )
    mock_target.fs.mount("c:", fs_win)

    mock_target.apply()

    return mock_target


@pytest.fixture
def target_unix(tmp_path: pathlib.Path, fs_unix: Filesystem) -> Target:
    return make_os_target(tmp_path, UnixPlugin, root_fs=fs_unix)


@pytest.fixture
def target_linux(tmp_path: pathlib.Path, fs_linux: Filesystem) -> Target:
    return make_os_target(tmp_path, LinuxPlugin, root_fs=fs_linux)


@pytest.fixture
def target_debian(tmp_path: pathlib.Path, fs_debian: Filesystem) -> Target:
    return make_os_target(tmp_path, DebianPlugin, root_fs=fs_debian)


@pytest.fixture
def target_redhat(tmp_path: pathlib.Path, fs_redhat: Filesystem) -> Target:
    return make_os_target(tmp_path, RedHatPlugin, root_fs=fs_redhat)


@pytest.fixture
def target_suse(tmp_path: pathlib.Path, fs_suse: Filesystem) -> Target:
    return make_os_target(tmp_path, SuSEPlugin, root_fs=fs_suse)


@pytest.fixture
def target_macos(tmp_path: pathlib.Path, fs_macos: Filesystem) -> Target:
    mock_target = make_os_target(tmp_path, MacOSPlugin, root_fs=fs_macos)

    version = absolute_path("_data/plugins/os/unix/bsd/darwin/macos/_os/SystemVersion.plist")
    fs_macos.map_file("/System/Library/CoreServices/SystemVersion.plist", version)

    system = absolute_path("_data/plugins/os/unix/bsd/darwin/macos/_os/preferences.plist")
    fs_macos.map_file("/Library/Preferences/SystemConfiguration/preferences.plist", system)

    return mock_target


@pytest.fixture
def target_ios(tmp_path: pathlib.Path, fs_ios: Filesystem) -> Target:
    return make_os_target(tmp_path, IOSPlugin, root_fs=fs_ios)


@pytest.fixture
def target_citrix(tmp_path: pathlib.Path, fs_bsd: VirtualFilesystem) -> Target:
    mock_target = next(make_mock_target(tmp_path))
    mock_target._os_plugin = CitrixPlugin

    mock_target.filesystems.add(fs_bsd)

    var_filesystem = VirtualFilesystem()
    var_filesystem.makedirs("/netscaler")
    var_filesystem.makedirs("/log")
    mock_target.filesystems.add(var_filesystem)

    flash_filesystem = VirtualFilesystem()
    flash_filesystem.map_dir("/", absolute_path("_data/plugins/os/unix/bsd/citrix/_os/flash"))
    mock_target.filesystems.add(flash_filesystem)

    mock_target.apply()
    return mock_target


@pytest.fixture
def target_android(tmp_path: pathlib.Path, fs_android: Filesystem) -> Target:
    return make_os_target(tmp_path, AndroidPlugin, root_fs=fs_android)


def add_win_user(hive_hklm: VirtualHive, hive_hku: VirtualHive, target_win: Target, sid: str, home: str) -> None:
    """Add a user to the provided Windows target."""

    profile_list_key_name = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList"
    try:
        profile_list_key = hive_hklm.key(profile_list_key_name)
    except RegistryKeyNotFoundError:
        profile_list_key = VirtualKey(hive_hklm, profile_list_key_name)

    try:
        hive_hklm.key(f"{profile_list_key_name}\\{sid}")
        raise ValueError(f"User with SID {sid} already exists!")
    except RegistryKeyNotFoundError:
        pass

    profile_key = VirtualKey(hive_hklm, f"{profile_list_key_name}\\{sid}")
    profile_key.add_value("ProfileImagePath", VirtualValue(hive_hklm, "ProfileImagePath", home))
    profile_list_key.add_subkey(sid, profile_key)

    hive_hklm.map_key(profile_list_key_name, profile_list_key)

    if sid != "S-1-5-18":
        target_win.registry.add_hive("HKEY_USERS", f"HKEY_USERS\\{sid}", hive_hku, TargetPath(target_win.fs, ""))


@pytest.fixture
def target_win_users(hive_hklm: VirtualHive, hive_hku: VirtualHive, target_win: Target) -> Target:
    """Minimal Windows target with a SYSTEM user and local user John."""
    add_win_user(hive_hklm, hive_hku, target_win, sid="S-1-5-18", home="%systemroot%\\system32\\config\\systemprofile")
    add_win_user(
        hive_hklm, hive_hku, target_win, sid="S-1-5-21-3263113198-3007035898-945866154-1002", home="C:\\Users\\John"
    )
    return target_win


@pytest.fixture
def target_win_tzinfo(hive_hklm: VirtualHive, target_win: Target) -> Target:
    tz_info_path = "SYSTEM\\ControlSet001\\Control\\TimeZoneInformation"
    tz_info = VirtualKey(hive_hklm, tz_info_path)
    tz_info.add_value("TimeZoneKeyName", "Easter Island Standard Time")

    eu_tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\W. Europe Standard Time"
    eu_tz_data = VirtualKey(hive_hklm, eu_tz_data_path)
    eu_tz_data.add_value("Display", "(UTC+01:00) Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna")
    eu_tz_data.add_value("Dlt", "W. Europe Daylight Time")
    eu_tz_data.add_value("Std", "W. Europe Standard Time")
    eu_tzi = bytes.fromhex("c4ffffff00000000c4ffffff00000a0000000500030000000000000000000300000005000200000000000000")
    eu_tz_data.add_value("TZI", eu_tzi)

    east_tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\Easter Island Standard Time"
    east_tz_data = VirtualKey(hive_hklm, east_tz_data_path)
    east_tz_data.add_value("Display", "(UTC-06:00) Easter Island")
    east_tz_data.add_value("Dlt", "Easter Island Daylight Time")
    east_tz_data.add_value("Std", "Easter Island Standard Time")
    east_tzi = bytes.fromhex("6801000000000000c4ffffff0000040006000100160000000000000000000900060001001600000000000000")
    east_tz_data.add_value("TZI", east_tzi)

    utc_tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\UTC"
    utc_tz_data = VirtualKey(hive_hklm, utc_tz_data_path)
    utc_tz_data.add_value("Display", "(UTC) Coordinated Universal Time")
    utc_tz_data.add_value("Dlt", "Coordinated Universal Time")
    utc_tz_data.add_value("Std", "Coordinated Universal Time")
    utc_tz_data.add_value("TZI", b"\x00" * 44)

    dynamic_dst = VirtualKey(hive_hklm, east_tz_data_path + "\\Dynamic DST")
    dynamic_dst.add_value("FirstEntry", 2019)
    dynamic_dst.add_value("LastEntry", 2019)
    tzi_2019 = bytes.fromhex("6801000000000000c4ffffff0000040006000100160000000000000000000900060001001600000000000000")
    dynamic_dst.add_value("2019", tzi_2019)
    east_tz_data.add_subkey("Dynamic DST", dynamic_dst)

    hive_hklm.map_key(tz_info_path, tz_info)
    hive_hklm.map_key(eu_tz_data_path, eu_tz_data)
    hive_hklm.map_key(east_tz_data_path, east_tz_data)
    hive_hklm.map_key(utc_tz_data_path, utc_tz_data)

    return target_win


@pytest.fixture
def target_win_tzinfo_legacy(hive_hklm: VirtualHive, target_win: Target) -> Target:
    tz_info_path = "SYSTEM\\ControlSet001\\Control\\TimeZoneInformation"
    tz_info = VirtualKey(hive_hklm, tz_info_path)
    tz_info.add_value("StandardName", "Paaseiland")

    east_tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\Easter Island Standard Time"
    east_tz_data = VirtualKey(hive_hklm, east_tz_data_path)
    east_tz_data.add_value("Display", "(UTC-06:00) Easter Island")
    east_tz_data.add_value("Dlt", "Easter Island Daylight Time")
    east_tz_data.add_value("Std", "Paaseiland")
    east_tzi = bytes.fromhex("6801000000000000c4ffffff0000040006000100160000000000000000000900060001001600000000000000")
    east_tz_data.add_value("TZI", east_tzi)

    hive_hklm.map_key(tz_info_path, tz_info)
    hive_hklm.map_key(east_tz_data_path, east_tz_data)

    return target_win


@pytest.fixture
def target_unix_users(target_unix: Target, fs_unix: Filesystem) -> Target:
    passwd = """
    root:x:0:0:root:/root:/bin/bash
    user:x:1000:1000:user:/home/user:/bin/bash
    +@ngtest:x:::::
    """
    fs_unix.map_file_fh("/etc/passwd", BytesIO(textwrap.dedent(passwd).encode()))
    return target_unix


@pytest.fixture
def target_linux_users(target_linux: Target, fs_linux: VirtualFilesystem) -> Target:
    passwd = """
    root:x:0:0:root:/root:/bin/bash
    user:x:1000:1000:user:/home/user:/bin/bash
    """
    fs_linux.map_file_fh("/etc/passwd", BytesIO(textwrap.dedent(passwd).encode()))
    return target_linux


@pytest.fixture
def target_macos_users(target_macos: Target, fs_macos: VirtualFilesystem) -> Target:
    dissect = absolute_path("_data/plugins/os/unix/bsd/darwin/macos/_os/dissect.plist")
    fs_macos.map_file("/var/db/dslocal/nodes/Default/users/_dissect.plist", dissect)

    test = absolute_path("_data/plugins/os/unix/bsd/darwin/macos/_os/test.plist")
    fs_macos.map_file("/var/db/dslocal/nodes/Default/users/_test.plist", test)

    fs_macos.makedirs("/Users/dissect")

    return target_macos


@pytest.fixture
def fs_docker() -> TarFilesystem:
    docker_tar = pathlib.Path(absolute_path("_data/plugins/apps/container/docker/docker.tgz"))
    fh = docker_tar.open("rb")
    return TarFilesystem(fh)


@pytest.fixture
def target_linux_docker(tmp_path: pathlib.Path, fs_docker: TarFilesystem) -> Target:
    mock_target = next(make_mock_target(tmp_path))
    mock_target._os_plugin = LinuxPlugin

    mock_target.filesystems.add(fs_docker)
    mock_target.fs.mount("/", fs_docker)
    mock_target.apply()
    return mock_target


class TargetUnixFactory:
    def __init__(self, tmp_path: pathlib.Path):
        self.tmp_path = tmp_path

    def new(self, hostname: str = "hostname") -> tuple[Target, VirtualFilesystem]:
        """Initialize a virtual unix target."""
        fs = VirtualFilesystem()

        fs.makedirs("var")
        fs.makedirs("etc")
        fs.map_file_fh("/etc/hostname", BytesIO(hostname.encode()))

        return make_os_target(self.tmp_path, UnixPlugin, root_fs=fs), fs


@pytest.fixture
def target_unix_factory(tmp_path: pathlib.Path) -> TargetUnixFactory:
    """This fixture returns a class that can instantiate a virtual unix targets from a blueprint. This can then be used
    to create a fixture for the source target and the desination target, without them 'bleeding' into each other."""
    return TargetUnixFactory(tmp_path)


@pytest.fixture(autouse=True)
def guarded_keychain() -> Iterator[None]:
    """This fixture clears the keychain from any previously added values."""
    keychain.KEYCHAIN.clear()
    yield
    keychain.KEYCHAIN.clear()



from __future__ import annotations

import io
import posixpath
import stat
import sys
from typing import IO, TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.filesystem import Filesystem, FilesystemEntry
    from dissect.target.helpers.fsutil import TargetPath

from dissect.target.exceptions import FilesystemError, SymlinkRecursionError
from dissect.target.helpers.polypath import abspath, normalize

try:
    # Up until CPython 3.12, pathlib._PathParents requires subclassing to inject the filesystem and flavour
    # into each parent path component. Since CPython 3.12, this is no longer necessary.
    # In CPython 3.13, _PathParents was moved to a different file, so this will result in an import error.
    # Since we no longer need it in CPython 3.13, we can just ignore the error.
    from pathlib import _PathParents

    class _DissectPathParents(_PathParents):
        __slots__ = ("_flavour", "_fs")

        def __init__(self, path: TargetPath):
            super().__init__(path)
            self._fs = path._fs
            self._flavour = path._flavour

        if sys.version_info >= (3, 10):

            def __getitem__(self, idx: int) -> TargetPath:
                result = super().__getitem__(idx)
                result._fs = self._fs
                result._flavour = self._flavour
                return result

        else:

            def __getitem__(self, idx: int) -> TargetPath:
                if idx < 0:
                    idx = len(self) + idx

                result = super().__getitem__(idx)
                result._fs = self._fs
                result._flavour = self._flavour
                return result

except ImportError:
    pass


class _DissectScandirIterator:
    """This class implements a ScandirIterator for dissect's scandir()

    The _DissectScandirIterator provides a context manager, so scandir can be called as:

    .. code-block:: python

        with scandir(path) as it:
            for entry in it
                print(entry.name)

    similar to os.scandir() behaviour since Python 3.6.
    """

    def __init__(self, iterator: Iterator[FilesystemEntry]):
        self._iterator = iterator

    def __del__(self) -> None:
        self.close()

    def __enter__(self) -> Iterator[FilesystemEntry]:
        return self._iterator

    def __exit__(self, *args, **kwargs) -> Literal[False]:
        return False

    def __iter__(self) -> Iterator[FilesystemEntry]:
        return self._iterator

    def __next__(self, *args) -> FilesystemEntry:
        return next(self._iterator, *args)

    def close(self) -> None:
        # close() is not defined in the various filesystem implementations. The
        # python ScandirIterator does define the interface however.
        pass


def scandir(path: TargetPath) -> _DissectScandirIterator:
    return _DissectScandirIterator(path.get().scandir())


def realpath(path: TargetPath, *, strict: bool = False) -> str:
    """Return the canonical path of the specified filename, eliminating any symbolic links encountered in the path."""
    filename = str(path)
    path, _ = _joinrealpath(path._fs, filename[:0], filename, strict, {})
    return abspath(path)


def isjunction(path: TargetPath) -> bool:
    """Return True if the path is a junction."""
    try:
        from dissect.target.filesystems.ntfs import NtfsFilesystemEntry
    except ImportError:
        return False

    entry = path.get()
    # Python's ntpath isjunction() only checks for mount point reparse tags
    return isinstance(entry, NtfsFilesystemEntry) and entry.dereference().is_mount_point()


# Join two paths, normalizing and eliminating any symbolic links
# encountered in the second path.
# NOTE: This is a copy of posixpath._joinrealpath with some small tweaks
def _joinrealpath(fs: Filesystem, path: str, rest: str, strict: bool, seen: dict[str, str]) -> tuple[str, bool]:
    if posixpath.isabs(rest):
        rest = rest[1:]
        path = "/"

    while rest:
        name, _, rest = rest.partition("/")
        if not name or name == ".":
            # current dir
            continue
        if name == "..":
            # parent dir
            if path:
                path, name = posixpath.split(path)
                if name == "..":
                    path = posixpath.join(path, "..", "..")
            else:
                path = ".."
            continue
        newpath = posixpath.join(path, name)
        try:
            st = fs.get(newpath).lstat()
        except FilesystemError:
            if strict:
                raise
            is_link = False
        else:
            is_link = stat.S_ISLNK(st.st_mode)
        if not is_link:
            path = newpath
            continue
        # Resolve the symbolic link
        if newpath in seen:
            # Already seen this path
            path = seen[newpath]
            if path is not None:
                # use cached value
                continue
            # The symlink is not resolved, so we must have a symlink loop.
            if strict:
                # Raise OSError(errno.ELOOP)
                raise SymlinkRecursionError(newpath)
            # Return already resolved part + rest of the path unchanged.
            return posixpath.join(newpath, rest), False
        seen[newpath] = None  # not resolved symlink
        path, ok = _joinrealpath(fs, path, normalize(fs.readlink(newpath)), strict, seen)
        if not ok:
            return posixpath.join(path, rest), False
        seen[newpath] = path  # resolved symlink

    return path, True


def io_open(
    path: TargetPath,
    mode: str = "rb",
    buffering: int = 0,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
) -> IO:
    """Open file and return a stream.

    Supports a subset of features of the real pathlib.open/io.open.

    Note: in contrast to regular Python, the mode is binary by default. Text mode
    has to be explicitly specified. Buffering is also disabled by default.
    """
    modes = set(mode)
    if modes - set("rbt") or len(mode) > len(modes):
        raise ValueError(f"invalid mode: {mode!r}")

    reading = "r" in modes
    binary = "b" in modes
    text = "t" in modes or "b" not in modes

    if "b" not in mode:
        encoding = encoding or "UTF-8"
        # CPython >= 3.10
        if hasattr(io, "text_encoding"):
            # Vermin linting needs to be skipped for this line as this is
            # guarded by an explicit check for availability.
            # novermin
            encoding = io.text_encoding(encoding)

    if not reading:
        raise ValueError("must be reading mode")
    if text and binary:
        raise ValueError("can't have text and binary mode at once")
    if binary and encoding is not None:
        raise ValueError("binary mode doesn't take an encoding argument")
    if binary and errors is not None:
        raise ValueError("binary mode doesn't take an errors argument")
    if binary and newline is not None:
        raise ValueError("binary mode doesn't take a newline argument")

    raw = path.get().open()
    result = raw

    line_buffering = False
    if buffering == 1 or (buffering < 0 and raw.isatty()):
        buffering = -1
        line_buffering = True
    if buffering < 0 or (text and buffering == 0):
        buffering = io.DEFAULT_BUFFER_SIZE
    if buffering == 0:
        if binary:
            return result
        raise ValueError("can't have unbuffered text I/O")

    buffer = io.BufferedReader(raw, buffering)
    result = buffer
    if binary:
        return result

    result = io.TextIOWrapper(buffer, encoding, errors, newline, line_buffering)
    result.mode = mode

    return result



#!/usr/bin/env python
from __future__ import annotations

import argparse
import itertools
import sys
from collections import deque
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

import structlog

from dissect.target.exceptions import FatalError, PluginNotFoundError, UnsupportedPluginError
from dissect.target.target import Target
from dissect.target.tools.dump.state import (
    DumpState,
    create_state,
    load_state,
    persisted_state,
)
from dissect.target.tools.dump.utils import (
    Compression,
    Serialization,
    cached_sink_writers,
)
from dissect.target.tools.utils import (
    FunctionDescriptor,
    configure_generic_arguments,
    configure_plugin_arguments,
    execute_function_on_target,
    find_and_filter_plugins,
    open_targets,
    process_generic_arguments,
    process_plugin_arguments,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from flow.record import Record

    from dissect.target.helpers.record import TargetRecordDescriptor

log = structlog.get_logger("dissect.target.tools.dump.run")


@dataclass
class RecordStreamElement:
    target: Target
    func: FunctionDescriptor
    record: Record
    end_pos: int | None = None
    sink_path: Path | None = None


def get_targets(targets: list[str]) -> Iterator[Target]:
    """Return a generator with :class:`Target` objects for provided paths."""
    yield from Target.open_all(targets)


def execute_function(
    target: Target, function: FunctionDescriptor, dry_run: bool, arguments: list[str]
) -> Iterator[TargetRecordDescriptor]:
    """Execute function ``function`` on provided target ``target`` and return a generator
    with the records produced.

    Only output type ``record`` is supported for plugin functions.
    """

    local_log = log.bind(func=function, target=target)
    local_log.debug("Function execution")

    if dry_run:
        print(f"  execute: {function.name} ({function.path})")
        return

    if function.output != "record":
        local_log.info(
            "Skipping target/func pair since its output type is not a record",
            target=target.path,
            func=function.name,
        )
        return
    try:
        _, target_attr = execute_function_on_target(target, function)
    except UnsupportedPluginError as e:
        local_log.error("Unsupported plugin for %s: %s", function.name, e.root_cause_str())
        local_log.debug("%s", function, exc_info=e)
        return
    except PluginNotFoundError:
        local_log.error("Cannot find plugin `%s`", function)
        return
    except FatalError as e:
        e.emit_last_message(local_log.error)
        sys.exit(1)
    except Exception as e:
        local_log.error("Exception while executing function %s (%s): %s", function.name, function.path, e)
        local_log.debug("", exc_info=e)
        local_log.debug("Function info: %s", function)
        return

    # no support for function-specific arguments
    result = target_attr() if callable(target_attr) else target_attr

    try:
        yield from result
    except Exception as e:
        local_log.error("Error occurred while processing output of %s.%s: %s", function.qualname, function.name, e)
        local_log.debug("", exc_info=e)
        return


def produce_target_func_pairs(
    targets: Iterable[Target],
    state: DumpState,
) -> Iterator[tuple[Target, FunctionDescriptor]]:
    """Return a generator with target and function pairs for execution.

    Target and function pairs that correspond to finished sinks in provided state ``state`` are skipped.
    """
    pairs_to_skip = set()
    if state:
        pairs_to_skip.update((str(sink.target_path), sink.func) for sink in state.finished_sinks)

    for target in targets:
        for func_def in find_and_filter_plugins(state.functions, target, state.excluded_functions):
            if state and (target.path, func_def.name) in pairs_to_skip:
                log.info(
                    "Skipping target/func pair since its marked as done in provided state",
                    target=target.path,
                    func=func_def.name,
                    state=state.path,
                )
                continue

            yield (target, func_def)
            state.mark_as_finished(target, func_def.name)


def execute_functions(
    target_func_stream: Iterable[tuple[Target, FunctionDescriptor]], dry_run: bool, arguments: list[str]
) -> Iterator[RecordStreamElement]:
    """Execute a function on a target for target / function pairs in the stream.

    Returns a generator of ``RecordStreamElement`` objects.
    """
    for target, func in target_func_stream:
        for record in execute_function(target, func, dry_run, arguments):
            yield RecordStreamElement(target=target, func=func, record=record)


def log_progress(stream: Iterable[Any], step_size: int = 1000) -> Iterator[Any]:
    """Log a number of items that went though the generator stream after
    every N element (N is configured in ``step_size``).
    """
    i = 0
    targets = set()
    sinks = set()
    for i, element in enumerate(stream, start=1):
        yield element

        if i % step_size == 0:
            log.info("Processing status", processed=i, last=element)

        targets.add(element.target)
        if element.sink_path:
            sinks.add(element.sink_path)

    log.info("Pipeline stats", elements=i, targets=len(targets), sinks=len(sinks))


def sink_records(
    record_stream: Iterable[RecordStreamElement],
    state: DumpState,
) -> Iterator[RecordStreamElement]:
    """Persist records from the stream into appropriate sinks, per serialization, compression and record type."""
    with cached_sink_writers(state) as write_element:
        for element in record_stream:
            write_element(element)
            yield element


def persist_processing_state(
    record_stream: Iterable[RecordStreamElement],
    state: DumpState,
) -> Iterator[RecordStreamElement]:
    """Keep track of the pipeline state in a persistent state object."""
    with persisted_state(state) as save_state_periodically:
        for element in record_stream:
            save_state_periodically()
            yield element


def configure_state(args: argparse.Namespace) -> DumpState | None:
    state = None if args.restart else load_state(output_dir=args.output)

    if state:
        log.info("Resuming state", output=args.output, state=state.path)

        if (
            state.serialization != args.serialization
            or state.compression != args.compression
            or state.functions != args.function
            or state.excluded_functions != args.excluded_functions
            or state.target_paths != args.targets
        ):
            log.error(
                "Configuration of the existing state conflicts with parameters provided",
                state_serialization=state.serialization,
                state_compression=state.compression,
                state_targets=state.target_paths,
                state_functions=state.functions,
                state_excluded_fucntions=state.excluded_functions,
                targets=args.targets,
                functions=args.function,
                serialization=args.serialization,
                compression=args.compression,
                state_path=state.path,
            )
            return None
    else:
        state = create_state(
            output_dir=args.output,
            target_paths=args.targets,
            functions=args.function,
            excluded_functions=args.excluded_functions,
            serialization=args.serialization,
            compression=args.compression,
        )
        log.info("New state created", restart=args.restart, state=state.path)

    return state


def execute_pipeline(
    state: DumpState,
    targets: Iterator[Target],
    dry_run: bool,
    arguments: list[str],
    limit: int | None = None,
) -> None:
    """Run the record generation, processing and sinking pipeline."""

    target_func_pairs_stream = produce_target_func_pairs(targets, state)
    record_stream = itertools.islice(execute_functions(target_func_pairs_stream, dry_run, arguments), limit)
    record_stream = sink_records(record_stream, state)
    record_stream = log_progress(record_stream)
    record_stream = persist_processing_state(record_stream, state)

    # exhaust the generator, executing all pipeline steps
    deque(record_stream, maxlen=0)

    log.info("Pipeline has finished")


def parse_arguments() -> tuple[argparse.Namespace, list[str]]:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
        add_help=False,
    )
    parser.add_argument("targets", metavar="TARGET", nargs="*", help="targets to load")

    configure_plugin_arguments(parser)

    parser.add_argument(
        "-c",
        "--compression",
        choices=[c.value for c in Compression if c is not Compression.NONE],
        type=Compression,
        help="compression method",
        default=Compression.NONE,
    )
    parser.add_argument(
        "--restart",
        action="store_true",
        help="restart the session and overwrite the state file if it exists",
    )
    parser.add_argument(
        "-s",
        "--serialization",
        choices=[s.value for s in Serialization],
        default=Serialization.JSONLINES,
        type=Serialization,
        help="serialization method",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=".",
        required=True,
        help="output directory",
    )
    parser.add_argument("--limit", type=int, help="limit number of records produced")

    configure_generic_arguments(parser)

    args, rest = parser.parse_known_args()
    process_generic_arguments(args)

    if not args.function and ("-h" in rest or "--help" in rest):
        parser.print_help()
        parser.exit(0)

    process_plugin_arguments(parser, args, rest)

    return args, rest


def main() -> None:
    args, rest = parse_arguments()

    try:
        state = configure_state(args)
        if state is None:
            # Error was already shown above, stopping execution
            return
        targets = open_targets(args)
        execute_pipeline(
            state=state,
            targets=targets,
            arguments=rest,
            dry_run=args.dry_run,
            limit=args.limit,
        )
    except Exception:
        log.exception("Exception while running the pipeline")
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime

from dissect.target.helpers import fsutil
from dissect.target.plugins.os.windows.notifications import (
    NOTIFICATIONS_DIR,
    NotificationsPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target

USER_DIR = "Users\\John"


def test_notifications_appdb(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    test_file = absolute_path("_data/plugins/os/windows/notifications/appdb.dat.v3.gz")
    appdb_file = fsutil.join(USER_DIR, NOTIFICATIONS_DIR, "appdb.dat")
    fs_win.map_file(appdb_file, test_file, compression="gzip")
    target_win_users.add_plugin(NotificationsPlugin)

    records_by_type = defaultdict(list)
    for record in target_win_users.notifications.appdb():
        record_type = record._desc.name
        records_by_type[record_type].append(record)

    records_by_type_chunk = defaultdict(lambda: defaultdict(list))
    for record_type, records in records_by_type.items():
        if record_type != "windows/notification/appdb":
            for record in records:
                chunk_num = record.chunk_num
                records_by_type_chunk[record_type][chunk_num].append(record)

    assert len(records_by_type.get("windows/notification/appdb", [])) == 1
    assert len(records_by_type.get("windows/notification/appdb/push", [])) == 1
    assert len(records_by_type.get("windows/notification/appdb/badge", [])) == 2
    assert len(records_by_type.get("windows/notification/appdb/tile", [])) == 48
    assert len(records_by_type.get("windows/notification/appdb/toast", [])) == 35

    appdb_record = records_by_type["windows/notification/appdb"][0]
    assert appdb_record.timestamp == datetime("2016-06-02T10:00:34.019495+00:00")
    assert appdb_record.chunk_version == 3
    assert appdb_record.next_notification_id == 517

    push_record = records_by_type_chunk["windows/notification/appdb/push"][8][0]
    assert push_record.push_ts1 == datetime("2016-06-19T07:37:38+00:00")
    assert push_record.push_ts2 == datetime("2016-05-20T07:41:45.883432+00:00")
    assert push_record.push_uri == (
        "https://db5.notify.windows.com/?token=AwYAAAA%2bwMdZymXtvB9uG3YbJZX4U"
        "CXwsLBJA7it1REPu58SjiQ8%2bnxg%2bfk8vKU%2bQQPG5ZglOuCq%2fkArOGxBJr9z7G"
        "K%2bQFwyrTcaOyptsKNF2f%2fllCPmmGwXsFAFjS%2fkdC678PQ%3d"
    )

    badge_record = records_by_type_chunk["windows/notification/appdb/badge"][3][0]
    assert badge_record.badge_ts is None
    assert badge_record.badge_id == 0x7A
    assert badge_record.badge_data.startswith("<badge")

    badge_record = records_by_type_chunk["windows/notification/appdb/badge"][10][0]
    assert badge_record.badge_ts is None
    assert badge_record.badge_id == 0x18
    assert badge_record.badge_data.startswith("<badge")

    tiles = records_by_type_chunk["windows/notification/appdb/tile"][0]
    tile_ids = [0x01F9, 0x01AF, 0x01B6, 0x01BC, 0x01F3]
    for tile_no, tile in enumerate(tiles):
        assert tile.arrival_time.year == 2016
        assert tile.expiry_time.year == 2016
        assert tile.id == tile_ids[tile_no]
        assert tile.tile_type == 3
        assert tile.index == tile_no
        assert tile.name == ""
        assert tile.content.startswith("<?xml")

    toasts = records_by_type_chunk["windows/notification/appdb/toast"][3]
    toast_ids = [0x00BF, 0x00C1, 0x00B7, 0x00C0]
    toast_name1 = ["0", "0", "0", "1"]
    for toast_num, toast in enumerate(toasts):
        assert toast.id == toast_ids[toast_num]
        assert toast.name1 in toast_name1[toast_num]
        assert toast.name2 == "MailGroup"


def test_notifications_wpndatabase(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    test_file = absolute_path("_data/plugins/os/windows/notifications/wpndatabase.db")
    wpndatabase_file = fsutil.join(USER_DIR, NOTIFICATIONS_DIR, "wpndatabase.db")
    fs_win.map_file(wpndatabase_file, test_file)

    target_win_users.add_plugin(NotificationsPlugin)

    records = list(target_win_users.notifications.wpndatabase())

    assert len(records) == 3



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.plugins.os.windows.registry import RegistryPlugin
from dissect.target.tools.reg import main as target_reg
from tests._utils import absolute_path

if TYPE_CHECKING:
    from _pytest.fixtures import FixtureRequest

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_win_reg(target_win: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_file("Windows/System32/config/SYSTEM", absolute_path("_data/tools/reg/SYSTEM.gz"), "gzip")
    target_win.apply()
    target_win.add_plugin(RegistryPlugin)
    return target_win


@pytest.mark.parametrize(
    ("provided_target", "provided_key", "arg_depth", "arg_length", "expected_output", "expected_log"),
    [
        # test invalid target
        ("target_default", "FOOBAR", 1, 100, "", "has no Windows Registry"),
        # test invalid key feedback
        ("target_win_reg", "FOOBAR", 1, 100, "", "Key 'FOOBAR' does not exist"),
        (
            "target_win_reg",
            "HKEY_LOCAL_MACHINE\\FOOBAR",
            1,
            100,
            "",
            "Key 'HKEY_LOCAL_MACHINE\\\\FOOBAR' does not exist",
        ),
        # test key value abbrevation
        (
            "target_win_reg",
            "HKEY_LOCAL_MACHINE\\SYSTEM\\ControlSet001\\Control\\ComputerName",
            1,
            100,
            "- 'ComputerName' 'DESKTOP-M3OSHQU'",
            "",
        ),
        (
            "target_win_reg",
            "HKEY_LOCAL_MACHINE\\SYSTEM\\ControlSet001\\Control\\ComputerName",
            1,
            10,
            "- 'ComputerName' 'DESKTOP-M...",
            "",
        ),
        # test class name printing
        (
            "target_win_reg",
            "HKLM\\SYSTEM\\ControlSet001\\Control\\Lsa\\Data",
            1,
            100,
            "+ 'Data' (2022-06-23 01:25:23.729778+00:00) (a282942c)",
            "",
        ),
    ],
)
def test_reg_output(
    provided_target: Target,
    provided_key: str,
    expected_output: str,
    expected_log: str,
    arg_depth: int,
    arg_length: int,
    capsys: pytest.CaptureFixture,
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    request: FixtureRequest,
) -> None:
    provided_target = request.getfixturevalue(provided_target)
    with patch("dissect.target.Target.open_all", return_value=[provided_target]), monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-reg", "foo", "--depth", str(arg_depth), "--length", str(arg_length), "--key", provided_key],
        )
        target_reg()
        out, err = capsys.readouterr()
        print(out, err)
        assert expected_output in out or expected_output in err
        assert expected_log in caplog.text



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator


class QemuChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields all QEMU domains from a KVM libvirt deamon."""

    __type__ = "qemu"

    def check_compatible(self) -> None:
        if not self.target.fs.path("/etc/libvirt/qemu").exists():
            raise UnsupportedPluginError("No libvirt QEMU installation found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for domain in self.target.fs.path("/etc/libvirt/qemu").glob("*.xml"):
            yield ChildTargetRecord(type=self.__type__, path=domain)



from __future__ import annotations

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.localeutil import normalize_language
from dissect.target.plugin import OperatingSystem, Plugin, export
from dissect.target.plugins.os.unix.locale import timezone_from_path


class LocalePlugin(Plugin):
    """Locale plugin for iOS targets."""

    def check_compatible(self) -> None:
        if not self.target.os == OperatingSystem.IOS:
            raise UnsupportedPluginError

    @export(property=True)
    def timezone(self) -> str | None:
        """Return the configured localtime of the iOS system."""
        if (localtime := self.target.fs.path("/private/var/db/timezone/localtime")).exists():
            return timezone_from_path(localtime.read_text().strip())
        return None

    @export(property=True)
    def language(self) -> list[str]:
        """Return the configured language(s) of the iOS system."""
        return list(map(normalize_language, self.target._os._config.GLOBAL["AppleLanguages"]))



from __future__ import annotations

import io
import stat
import struct
import zipfile
import zlib
from typing import TYPE_CHECKING, BinaryIO

import defusedxml.ElementTree as ET
from dissect.util import ts

from dissect.target import filesystem
from dissect.target.helpers import fsutil, record, regutil
from dissect.target.loader import Loader
from dissect.target.plugin import OSPlugin, export
from dissect.target.plugins.os.windows.registry import RegistryPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path
    from xml.etree.ElementTree import Element

    from typing_extensions import Self

    from dissect.target.target import Target

EXTENSIONS = ("upr", "upf", "upr.zip", "upf.zip")

PATH_REPLACEMENTS = {
    r"%appdata%": "users/{user}/appdata/roaming",
    r"%localappdata%": "users/{user}/appdata/local",
    r"%localappdata%low": "users/{user}/appdata/locallow",
}


def find_pwr_dir(path: Path) -> Path | None:
    if not path.is_dir():
        return None

    pwr_dirs = [
        "Pwrmenu2",
        "PwrmenuPrd",
        "PwrmenuGAT",
    ]
    for p in pwr_dirs:
        pwrdir = path.joinpath(p)
        if not pwrdir.exists():
            continue

        userpref = pwrdir.joinpath("UserPref")
        if userpref.exists():
            for userpref_files in userpref.iterdir():
                if userpref_files.name.endswith(EXTENSIONS):
                    return pwrdir

                epath = userpref.joinpath(userpref_files)
                if epath.is_dir() and any(usrset_file.name.endswith(EXTENSIONS) for usrset_file in epath.iterdir()):
                    return pwrdir
    return None


class ResLoader(Loader):
    @staticmethod
    def detect(path: Path) -> bool:
        return find_pwr_dir(path) is not None

    def map(self, target: Target) -> None:
        path = find_pwr_dir(self.absolute_path)
        user_inf_path = path.joinpath("user.inf")

        vfs = filesystem.VirtualFilesystem(case_sensitive=False)
        target.filesystems.add(vfs)

        regflex = regutil.RegFlex()

        username = user_inf_path.read_text().strip() if user_inf_path.exists() else path.parts[-2]

        userpref = path.joinpath("UserPref")
        for fpath in userpref.rglob("*"):
            if not fpath.is_file():
                continue

            if fpath.suffix == ".zip":
                try:
                    zfile = zipfile.ZipFile(fpath, "r")
                except zipfile.BadZipFile:
                    pass

                fh = io.BytesIO(zfile.open(zfile.namelist()[0]).read())
            else:
                fh = fpath.open("rb")

            if "upr" in fpath.name:
                upr = UPR(fh)
                if upr.buf[:7] != b"RESZLIB":
                    continue
                regflex.map_definition(upr.open())

            if "upf" in fpath.name:
                try:
                    upf = UPF(fh)
                except Exception:
                    continue

                for fileobj in upf.files():
                    fpath = fileobj.path.replace("{user}", username)
                    fentry = ResFile(vfs, fpath, fileobj)
                    vfs.map_file_entry(fentry.path, fentry)

        target.props["username"] = username
        target._os_plugin = ResOSPlugin.create(target, vfs)
        target.add_plugin(RegistryPlugin, check_compatible=False)

        for name, hive in regflex.hives.items():
            path = name
            if name == "HKEY_CURRENT_USER":
                # name = 'S-0'
                path = "HKEY_USERS\\S-0"
            target.registry.add_hive(name, path, hive, fsutil.TargetPath(target.fs, "UPR"))


class ResFile(filesystem.VirtualFile):
    entry: File

    def __init__(self, fs: filesystem.VirtualFilesystem, path: str, entry: File, **kwargs):
        super().__init__(fs, path, entry)

    def stat(self) -> fsutil.stat_result:
        return fsutil.stat_result(
            [
                stat.S_IFREG,
                0,
                0,
                0,
                0,
                0,
                self.entry.size,
                ts.to_unix(self.entry.timestamps[2]),
                ts.to_unix(self.entry.timestamps[1]),
                ts.to_unix(self.entry.timestamps[0]),
            ]
        )

    def lstat(self) -> fsutil.stat_result:
        return self.stat()

    def open(self) -> BinaryIO:
        return self.entry.open()


class ResOSPlugin(OSPlugin):
    @classmethod
    def detect(cls, target: Target) -> bool:
        return True

    @classmethod
    def create(cls, target: Target, sysvol: filesystem.Filesystem) -> Self:
        target.fs.case_sensitive = False
        target.fs.mount("sysvol", sysvol)
        return cls(target)

    @export(property=True)
    def hostname(self) -> str:
        return self.target.props["username"]

    @export(property=True)
    def ips(self) -> list[str]:
        return []

    @export(property=True)
    def version(self) -> None:
        return None

    @export
    def users(self) -> Iterator[record.WindowsUserRecord]:
        yield record.WindowsUserRecord(
            sid="S-0",
            name=self.hostname,
            home=f"sysvol/users/{self.hostname}",
            _target=self.target,
        )

    @export(property=True)
    def os(self) -> str:
        return "windows"


class UPR:
    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.buf = fh.read()

    def open(self) -> io.StringIO:
        buf = self.buf
        if buf[:7] == b"RESZLIB":
            # Skip header
            buf = zlib.decompress(buf[11:])
        return io.StringIO(buf.decode("latin1"))


class UPF:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        self._meta = self._read_metadata()
        self.metadata = ET.fromstring(self._meta)

        self._folders = {}
        self._files = {}

        self._parse()

    def folders(self) -> list[Folder]:
        return list(self._folders.values())

    def files(self) -> list[File]:
        return list(self._files.values())

    def _read_metadata(self) -> str:
        needle = b"R\x00E\x00S\x00Z\x00L\x00I\x00B\x00"

        if (offset := reverse_search(self.fh, needle)) == -1:
            raise ValueError("Invalid UPF file")

        # Skip header
        self.fh.seek(offset + 22)
        return zlib.decompress(self.fh.read()).decode("utf-16-le")

    def _parse(self) -> None:
        folders = self.metadata.find("folders")
        if folders is not None:
            for elem in folders.findall("folder"):
                self._iter_folder(elem)

        files = self.metadata.find("files")
        if files is not None:
            for elem in files:
                res_file = File(self, elem)
                self._files[res_file.guid] = res_file

    def _iter_folder(self, elem: Element, parent: Folder | None = None) -> None:
        folder = Folder(self, elem, parent=parent)
        self._folders[folder.guid] = folder

        for child in elem.findall("folder"):
            self._iter_folder(child, folder)


class File:
    def __init__(self, upf: UPF, elem: Element):
        self.upf = upf
        self.name = elem.find("name").text
        self.size = int(elem.find("filelen").text)
        self.offset = int(elem.find("offset").text)
        self.packed_size = int(elem.find("len").text)
        self.compressed = elem.find("compressed").text == "yes"

        timestamps = struct.unpack(">3Q", bytes.fromhex(elem.find("timestamp").text))
        self.timestamps = [ts.wintimestamp(t) for t in timestamps]

        self.guid = elem.get("guid")
        self.folder = self.upf._folders.get(elem.get("folderguid"), None)

    def __repr__(self) -> str:
        return f"<File path={self.path} size={self.size} timestamps={[str(t) for t in self.timestamps]}>"

    @property
    def path(self) -> str:
        if not self.folder:
            return self.name
        return f"{self.folder.path}/{self.name}"

    def open(self) -> io.BytesIO:
        fh = self.upf.fh
        offset = self.offset - 1
        if self.compressed:
            # Skip header
            fh.seek(offset + 11)
            buf = zlib.decompress(fh.read(self.packed_size))
        else:
            fh.seek(offset)
            buf = fh.read(self.packed_size)

        return io.BytesIO(buf)


class Folder:
    def __init__(self, upf: UPF, elem: Element, parent: Folder | None = None):
        self.upf = upf
        self.name = elem.find("name").text
        self.guid = elem.get("guid")
        self.parent = parent

    def __repr__(self) -> str:
        return f"<Folder path={self.path}>"

    @property
    def path(self) -> str:
        name = PATH_REPLACEMENTS.get(self.name.lower(), self.name)
        if self.parent:
            return f"{self.parent.path}/{name}"
        return name


def reverse_search(fh: BinaryIO, needle: bytes) -> int:
    BLOCK_SIZE = 1024 * 1024 * 64

    fh.seek(0, io.SEEK_END)
    size = fh.tell()
    offset = size

    while offset > 0:
        read_size = min(offset, BLOCK_SIZE)
        offset -= read_size

        fh.seek(offset)
        buf = fh.read(read_size)

        pos = buf.find(needle)
        if pos == -1:
            continue

        return offset + pos

    return -1



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.hypervisor.disk import c_qcow2, qcow2

from dissect.target.container import Container

if TYPE_CHECKING:
    from pathlib import Path


class QCow2Container(Container):
    __type__ = "qcow2"

    def __init__(
        self,
        fh: BinaryIO | Path,
        data_file: BinaryIO | None = None,
        backing_file: BinaryIO | None = None,
        *args,
        **kwargs,
    ):
        self.qcow2 = qcow2.QCow2(fh, data_file, backing_file)
        self.stream = self.qcow2.open()

        super().__init__(fh, self.qcow2.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return fh.read(4) == c_qcow2.QCOW2_MAGIC_BYTES

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return path.suffix.lower() == ".qcow2"

    def read(self, length: int) -> bytes:
        return self.stream.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.stream.seek(offset, whence)

    def tell(self) -> int:
        return self.stream.tell()

    def close(self) -> None:
        self.stream.close()



from __future__ import annotations

import datetime
import re
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.helpers.utils import year_rollover_helper
from dissect.target.plugin import Plugin, alias, export
from dissect.target.plugins.os.unix.log.helpers import (
    RE_LINE,
    RE_TS,
    is_iso_fmt,
    iso_readlines,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

MessagesRecord = TargetRecordDescriptor(
    "linux/log/messages",
    [
        ("datetime", "ts"),
        ("string", "service"),
        ("varint", "pid"),
        ("string", "message"),
        ("path", "source"),
    ],
)

DEFAULT_TS_LOG_FORMAT = "%b %d %H:%M:%S"
RE_CLOUD_INIT_LINE = re.compile(
    r"^(?P<ts>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3}) - (?P<service>.*)\[(?P<log_level>\w+)\]\: (?P<message>.*)$"
)


class MessagesPlugin(Plugin):
    """Unix messages log plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.log_files = set(self._find_log_files())

    def _find_log_files(self) -> Iterator[Path]:
        log_dirs = ["/var/log/", "/var/log/installer/"]
        file_globs = ["syslog*", "messages*", "cloud-init.log*"]
        for log_dir in log_dirs:
            for glob in file_globs:
                yield from self.target.fs.path(log_dir).glob(glob)

    def check_compatible(self) -> None:
        if not self.log_files:
            raise UnsupportedPluginError("No log files found")

    @alias("syslog")
    @export(record=MessagesRecord)
    def messages(self) -> Iterator[MessagesRecord]:
        """Return contents of /var/log/messages*, /var/log/syslog* and cloud-init logs.

        Due to year rollover detection, the log contents could be returned in reversed or mixed chronological order.

        The messages log file holds information about a variety of events such as the system error messages, system
        startups and shutdowns, change in the network configuration, etc. Aims to store valuable, non-debug and
        non-critical messages. This log should be considered the "general system activity" log.

        References:
            - https://geek-university.com/linux/var-log-messages-file/
            - https://www.geeksforgeeks.org/file-timestamps-mtime-ctime-and-atime-in-linux/
            - https://cloudinit.readthedocs.io/en/latest/development/logging.html#logging-command-output
        """
        target_tz = self.target.datetime.tzinfo

        for log_file in self.log_files:
            if "cloud-init" in log_file.name:
                yield from self._parse_cloud_init_log(log_file, target_tz)
                continue

            if is_iso_fmt(log_file):
                iterable = iso_readlines(log_file)

            else:
                iterable = year_rollover_helper(log_file, RE_TS, DEFAULT_TS_LOG_FORMAT, target_tz)

            for ts, line in iterable:
                match = RE_LINE.search(line)

                if not match:
                    self.target.log.warning("Unable to parse message line in %s: %r", log_file, line)
                    continue

                yield MessagesRecord(
                    ts=ts,
                    **match.groupdict(),
                    source=log_file,
                    _target=self.target,
                )

    def _parse_cloud_init_log(
        self, log_file: Path, tzinfo: datetime.tzinfo | None = datetime.timezone.utc
    ) -> Iterator[MessagesRecord]:
        """Parse a cloud-init.log file.

        Lines are structured in the following format:
        ``YYYY-MM-DD HH:MM:SS,000 - dhcp.py[DEBUG]: Received dhcp lease on IFACE for IP/MASK``

        NOTE: ``cloud-init-output.log`` files are not supported as they do not contain structured logs.

        Args:
            ``log_file``: path to cloud-init.log file.

        Returns: ``MessagesRecord``
        """

        ts_fmt = "%Y-%m-%d %H:%M:%S,%f"

        with open_decompress(log_file, "rt") as fh:
            for line in fh:
                if not (line := line.strip()):
                    continue

                if not (match := RE_CLOUD_INIT_LINE.match(line)):
                    self.target.log.warning("Could not match cloud-init log line in file: %s", log_file)
                    self.target.log.debug("No match for line '%s'", line)
                    continue

                values = match.groupdict()

                # Actual format is ``YYYY-MM-DD HH:MM:SS,000`` (asctime with milliseconds) but python has no strptime
                # operator for 3 digit milliseconds, so we convert and pad to six digit microseconds.
                # https://github.com/canonical/cloud-init/blob/main/cloudinit/log/loggers.py#DEFAULT_LOG_FORMAT
                # https://docs.python.org/3/library/logging.html#asctime
                raw_ts, _, milliseconds = values["ts"].rpartition(",")
                raw_ts += "," + str(int(milliseconds) * 1000).zfill(6)

                try:
                    ts = datetime.datetime.strptime(raw_ts, ts_fmt).replace(tzinfo=tzinfo)

                except ValueError as e:
                    self.target.log.warning("Timestamp '%s' does not match format '%s'", raw_ts, ts_fmt)
                    self.target.log.debug("", exc_info=e)
                    ts = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, tzinfo=datetime.timezone.utc)

                yield MessagesRecord(
                    ts=ts,
                    service=values["service"],
                    pid=None,
                    message=values["message"],
                    source=log_file,
                    _target=self.target,
                )
